import pandas as _pd
import os as _os
import json as _json
import numpy as _np
import requests as _requests
from io import BytesIO as _BytesIO
from zipfile import ZipFile as _ZipFile
import warnings as _warnings
from io import BytesIO
from pandas.errors import ParserError


def get_gis(url="https://www.eia.gov/maps/map_data/CrudeOil_Pipelines_US_EIA.zip"):
    """
    Returns a SpatialPointsDataFrame from a shapefile URL. Examples with EIA and Government of Alberta

    US Energy Information Agency:
    - EIA crude pipelines : https://www.eia.gov/maps/map_data/CrudeOil_Pipelines_US_EIA.zip
    - EIA Refinery Map : https://www.eia.gov/maps/map_data/Petroleum_Refineries_US_EIA.zip
    - EIA Products Pipelines : https://www.eia.gov/maps/map_data/PetroleumProduct_Pipelines_US_EIA.zip
    - EIA Products Terminals : https://www.eia.gov/maps/map_data/PetroleumProduct_Terminals_US_EIA.zip
    - EIA NG Pipelines : https://www.eia.gov/maps/map_data/NaturalGas_InterIntrastate_Pipelines_US_EIA.zip
    - EIA NG Storage : https://www.eia.gov/maps/map_data/PetroleumProduct_Terminals_US_EIA.zip
    - EIA NG Hubs : https://www.eia.gov/maps/map_data/NaturalGas_TradingHubs_US_EIA.zip
    - EIA LNG Terminals : https://www.eia.gov/maps/map_data/Lng_ImportExportTerminals_US_EIA.zip

    Alberta Oil Sands, Petroleum and Natural Gas
    - AB : https://gis.energy.gov.ab.ca/GeoviewData/OS_Agreements_Shape.zip

    Parameters
    ----------
    url : str
        URL of the zipped shapefile

    Return
    ------
    Returns geopandas object

    Examples
    --------
    >>> import risktools as rt
    >>> df = rt.data.get_gis("https://www.eia.gov/maps/map_data/CrudeOil_Pipelines_US_EIA.zip")
    """

    try:
        import geopandas as _geopandas
    except:
        raise ImportError("Geopandas not installed. Please install before running")

    try:
        from fiona.io import ZipMemoryFile as _ZMF
    except:
        raise ImportError("Fiona not installed. Please install before running")

    fn = _requests.get(url)

    # useful for when there are multiple directories or files. Takes first shape file
    for ff in _ZipFile(_BytesIO(fn.content)).namelist():
        if ff[-4:] == ".shp":
            shp_file = ff
            break

    zf = _ZMF(fn.content)
    shp = zf.open(shp_file)

    return _geopandas.GeoDataFrame.from_features(shp, crs=shp.crs)


def get_names():
    """
    return valid names for the open_data() function.

    Returns
    -------
    List of strings

    Examples
    --------
    >>> import risktools as rt
    >>> rt.data.get_names()
    """
    return list(_file_actions.keys())


def open_data(nm):
    """
    Function used to return built-in datasets from risktools. To get a list of valid datasets, use the get_names() function.

    Parameters
    ----------
    nm : str
        Name of dataset to return

    Returns
    -------
    Varies by data requests

    Examples
    --------
    >>> import risktools as rt
    >>> rt.data.open_data('crudeOil')
    """
    fn = ""
    path = _os.path.dirname(__file__)
    try:
        fn = _file_actions[nm]["file"]
    except ValueError:
        print(f"{nm} is not a valid file name to open")

    fp = _os.path.join(_path, fn)

    try:
        df = _file_actions[nm]["load_func"](fp)
    except:
        _warnings.warn(f"File actions for {nm} not defined. Running default behavior.")
        df = _load_data(_os.path.join(_path, f"{nm}.json"))

    if isinstance(df, _pd.DataFrame):
        # convert "." to "_" in column names
        df.columns = df.columns.str.replace(".", "_", regex=False)

    # convert datetime fields
    if _file_actions[nm]["date_fields"] is not None:
        for d in _file_actions[nm]["date_fields"]:
            df[d] = _pd.to_datetime(df[d])

    if "index" in _file_actions[nm].keys():
        df = df.set_index(_file_actions[nm]["index"]).sort_index()
        if len(df.columns) < 2:
            df = df.iloc[:, 0]

    return df


def _norm_df(fn):

    with open(fn, mode="r") as file:
        tmp = _json.load(file)

    df = dict()

    for key in tmp.keys():
        
        # if structure of dict is {key : DataFrame}
        try: 
            df[key] = _pd.DataFrame.from_records(tmp[key])
            # df[key] = _try_dates(df[key])
        except:
            pass

        # if structure of dict is {first_key : {second_key : DataFrame}}
        # return {first_key : DataFrame} where second_key is a column
        # of the DataFrame
        try:    
            cf = _pd.DataFrame()
            for sec_key in tmp[key].keys():
                tf = _pd.DataFrame.from_records(tmp[key][sec_key])
                tf.columns = tf.columns.str.replace("\.+", "_", regex=True)
                cols = list(tf.columns)
                tf['assay'] = sec_key
                tf = tf[['assay', *cols]]
                cf = _pd.concat([cf, tf], axis=0)
            df[key] = cf
            # df[key] = _try_dates(df[key])

        except:
            pass
    
    return df


def _load_data(fn):

    with open(fn) as f:
        dd = _json.load(f)

    try:
        df = _pd.DataFrame.from_records(dd)
        df = _try_dates(df)
        df.columns = df.columns.str.replace("\.+", "_", regex=True)
        return df
    except:
        pass

    for key in dd.keys():

        try:
            dd[key] = _pd.DataFrame.from_records(dd[key])
            dd[key].columns = dd[key].columns.str.replace("\.+", "_", regex=True)
        except:
            pass

        try:
            dd[key] = _pd.DataFrame(dd[key])
            dd[key].columns = dd[key].columns.str.replace("\.+", "_", regex=True)
        except:
            pass

        if isinstance(dd[key], _pd.DataFrame) == True:
            dd[key] = _try_dates(dd[key])

        # try:
        #     dd[key] = _np.array(dd[key])
        # except:
        #     pass
    # dd.columns = dd.columns.str.replace("\.+", "_", regex=True)
    return dd


def _try_dates(df):
    for c in df.columns[df.dtypes == "object"]:  # don't cnvt num
        try:
            df[c] = _pd.to_datetime(df[c])
        except (ParserError, ValueError):  # Can't cnvrt some
            pass  # ...so leave whole column as-is unconverted

    return df


def _read_curves(fn):
    # open SwapCurve files (based on R S3 class Discount Curves) and convert to
    # dictionary with nested arrays, dictionaries and one nested dataframe

    with open(fn) as f:
        dd = _json.load(f)

    for key in dd.keys():
        if isinstance(dd[key], list) == True:
            dd[key] = _np.array(dd[key])

        if isinstance(dd[key], dict) == True:
            # print(dd[key])
            for k in dd[key].keys():
                dd[key][k] = _np.array(dd[key][k])

                if len(dd[key][k]) == 1:
                    dd[key][k] = dd[key][k][0]


    dd["table"] = _pd.DataFrame.from_records(dd["table"])

    dd["table"]["date"] = _pd.to_datetime(dd["table"]["date"])  # , utc=True, unit="D"

    return dd


def _read_dict(fn):
    with open(fn) as f:
        dd = _json.load(f)

    return dd


_file_actions = {
    "crudeOil": {
        "file": "crudeOil.json",
        "date_fields": None,
        "load_func": _norm_df,
    },
    "cushing": {
        "file": "cushing.json",
        "date_fields": None,
        "load_func": _load_data,
    },
    "dflong": {
        "file": "dflong.json",
        "date_fields": ["date"],
        "load_func": _pd.read_json,
        "index": ["series", "date"],
    },
    "dfwide": {
        "file": "dfwide.json",
        "date_fields": ["date"],
        "load_func": _pd.read_json,
        "index": "date",
    },
    "eiaStocks": {
        "file": "eiaStocks.json",
        "date_fields": ["date"],
        "load_func": _pd.read_json,
    },
    "eiaStorageCap": {
        "file": "eiaStorageCap.json",
        "date_fields": ["date"],
        "load_func": _pd.read_json,
    },
    "eurodollar": {
        "file": "eurodollar.json",
        "date_fields": None,
        "load_func": _load_data,
    },
    "expiry_table": {
        "file": "expiry_table.json",
        "date_fields": [
            "Last_Trade",
            "First_Notice",
            "First_Delivery",
            "Last_Delivery",
        ],
        "load_func": _pd.read_json,
    },
    "fizdiffs": {
        "file": "fizdiffs.json",
        "date_fields": ["date"],
        "load_func": _pd.read_json,
    },
    "futuresRef": {
        "file": "futuresRef.json",
        "date_fields": None,
        "load_func": _norm_df,
    },
    "fxfwd": {
        "file": "fxfwd.json",
        "date_fields": None,
        "load_func": _load_data,
    },
    "holidaysOil": {
        "file": "holidaysOil.json",
        "date_fields": ["value"],
        "load_func": _pd.read_json,
    },
    "ohlc": {
        "file": "ohlc.json",
        "date_fields": ["date"],
        "load_func": _pd.read_json,
    },
    "planets": {
        "file": "planets.json",
        "date_fields": None,
        "load_func": _pd.read_json,
    },
    "refineryLPdata": {
        "file": "refineryLPdata.json",
        "date_fields": None,
        "load_func": _pd.read_json,
    },
    "tickers_eia": {
        "file": "tickers_eia.json",
        "date_fields": None,
        "load_func": _pd.read_json,
    },
    "tradeCycle": {
        "file": "tradeCycle.json",
        "date_fields": ["flowmonth", "trade_cycle_end"],
        "load_func": _pd.read_json,
    },
    "tradeHubs": {
        "file": "tradeHubs.json",
        "date_fields": None,
        "load_func": _pd.read_json,
    },
    "tradeprocess": {
        "file": "tradeprocess.json",
        "date_fields": ["date"],
        "load_func": _pd.read_json,
    },
    "wti_swap": {
        "file": "wtiSwap.json",
        "date_fields": ["date"],
        "load_func": _pd.read_json,
    },
    "stocks": {
        "file": "stocks.json",
        "date_fields": None,
        "load_func": _load_data,
    },
    "tsQuotes": {
        "file": "tsQuotes.json",
        "date_fields": None,
        "load_func": _pd.read_json,
    },
    "usSwapCurves": {
        "file": "usSwapCurves.json",
        "date_fields": None,
        "load_func": _read_curves,
    },
    "usSwapCurvesPar": {
        "file": "usSwapCurvesPar.json",
        "date_fields": None,
        "load_func": _read_curves,
    },
}


_path = _os.path.dirname(__file__)


if __name__ == "__main__":
    pass



from ._charts import *
from ._pa import *
from ._swap import *
from ._sims import *
from ._main_functions import *
from ._multivariate import *
from .extensions import *

# from .data import get_gis
from ._refineryLP import *
from ._cullenfrey import describe_distribution

#####################################################################
# TODO
# * Add legend to chart_eia_sd function
#####################################################################


# if __name__ == "__main__":
#     _np.random.seed(42)
#     import json

#     with open("../../user.json") as jfile:
#         userfile = jfile.read()

#     up = json.loads(userfile)

#     username = up["m*"]["user"
#     password = up["m*"]["pass"]

#     print(
#         chart_spreads(
#             [("@HO4H", "@HO4J", "2014"), ("@HO9H", "@HO9J", "2019")],
#             200,
#             username,
#             password,
#             feed="CME_NymexFutures_EOD",
#             output="chart",
#             start_dt="2012-01-01",
#         )
#     )

# sp = simOU()
# print(fitOU(sp))
# print(sp)

# print(bond(output='price'))
# print(bond(output='duration'))
# print(bond(output='df'))

# dflong = data.open_data('dflong')
# ret = returns(df=dflong, ret_type="abs", period_return=1, spread=True).iloc[:,0:2]
# roll_adjust(df=ret, commodity_name="cmewti", roll_type="Last_Trade").head(50)

# from pandas_datareader import data, wb
# from datetime import datetime
# df = data.DataReader(["SPY","AAPL"],  "yahoo", datetime(2000,1,1), datetime(2012,1,1))
# df = df.pct_change()
# df = df.asfreq('B')

# R = df[('Adj Close','SPY')]

# rt.returns(df = rt.data.open_data('dflong'), ret_type = "rel", period_return = 1, spread = False)
# rt.returns(df = rt.data.open_data('dflong'), ret_type = "log", period_return = 1, spread = True)

# print(y)

# print(_check_ts(R.dropna(), scale=252))

# print(trade_stats(df[('Adj Close','SPY')]))
# print(trade_stats(df['Adj Close']))

# tt = _sr(df['Adj Close'], Rf=0, scale=252)
# print(tt)

# print(drawdowns(df['Adj Close']))
# print(drawdowns(df[('Adj Close','SPY')]))

# rs = find_drawdowns(df['Adj Close'])
# print(rs['SPY']['peaktotrough'])
# print(find_drawdowns(df[('Adj Close','SPY')]))

# print(sharpe_ratio_annualized(df['Adj Close']))
# print(sharpe_ratio_annualized(df[('Adj Close','SPY')]))

# print(omega_sharpe_ratio(df['Adj Close'],MAR=0))
# print(upside_risk(df['Adj Close'], MAR=0))
# print(upside_risk(df[('Adj Close','SPY')], MAR=0))
# print(upside_risk(df['Adj Close'], MAR=0, stat='variance'))
# print(upside_risk(df['Adj Close'], MAR=0, stat='potential'))

# print(downside_deviation(df[('Adj Close','SPY')], MAR=0))
# print(downside_deviation(df['Adj Close'], MAR=0))

# print(return_cumulative(r=df['Adj Close'], geometric=True))
# print(return_cumulative(r=df['Adj Close'], geometric=False))

# print(return_annualized(r=df[('Adj Close','SPY')], geometric=True))
# print(return_annualized(r=df[('Adj Close','SPY')], geometric=False))

# print(return_annualized(r=df['Adj Close'], geometric=True))
# print(return_annualized(r=df['Adj Close'], geometric=False))

# print(sd_annualized(x=df[('Adj Close','SPY')]))
# print(sd_annualized(x=df['Adj Close']))

####################
# R code for testing
# library(timetk)
# library(tidyverse)
# library(PerformanceAnalytics)
# df <- tq_get('SPY', from='2000-01-01', to='2012-01-01')
# df <- df %>% dplyr::mutate(adjusted = adjusted/lag(adjusted)-1)
# R = tk_xts(df, select=adjusted, date_var=date)
# DownsideDeviation(R, MAR=0)
# UpsideRisk(R, 0)
# OmegaSharpeRatio(R, 0)



from . import data

import pandas as _pd
import numpy as _np
from scipy import interpolate as _interpolate
from ._morningstar import *

import pandas_datareader as _pdr
import plotly.express as _px

us_swap = data.open_data("usSwapCurves")


def custom_date_range(start, end, freq):
    if freq == "M":
        mul = 1
    elif freq == "Q":
        mul = 3
    elif (freq == "H") | (freq == "6M"):
        mul = 6
    elif freq == "Y":
        mul = 12
    else:
        raise ValueError("freq is not in ['M','Q','H' or 'Y']")

    dr = [_pd.to_datetime(start)]

    i = 1
    while dr[-1] < _pd.to_datetime(end):
        dr.append(dr[0] + _pd.DateOffset(months=mul * i))
        i += 1

    return _pd.Index(dr)


def swap_irs(
    trade_date=None,
    eff_date=None,
    mat_date=None,
    notional=1000000,
    pay_rec="rec",
    fixed_rate=0.05,
    float_curve=None,
    reset_freq="Q",
    disc_curve=None,
    days_in_year=360,
    convention="act",
    bus_calendar="NY",  # not implemented
    output="price",
):
    """
    Commodity swap pricing from exchange settlement

    Parameters
    ----------
    trade_date : Timestamp | str
        The date of the trade. If None, defaults to today(). By default None.
    eff_date : Timestamp | str
        The effective date of the swap. If None, defaults to trade_date + 2 days. By default None.
    mat_date : Timestamp | str
        The maturity date of the swap. If None, defaults to eff_date + 2 years.
    notional : long int
        Numeric value of notional. Defaults to 1,000,000.
    pay_rec : str
        "pay" for pyement (positive pv) or "rec" for receivables (negative pv).
    fixed_rate : float
        fixed interest rate. Defaults to 0.05.
    float_curve : DataFrame | Dict
        DataFrame of interest rate curves with columns 'times' and 'discounts'. If None, defaults to rt.data.open_data("usSwapCurves")
        which is a dictionary based on the R object DiscountCurve. Column times is year fractions going forward in time.
        So today is 0, tomorrow is 1/365, a year from today is 1 etc...
    reset_freq : str
        Pandas/datetime Timestamp frequency (allowable values are 'M', 'Q', '6M', or 'Y')
    disc_curve : DataFrame | Dict
        DataFrame of interest rate curves with columns 'times' and 'discounts'. If None, defaults to rt.data.open_data("usSwapCurves")
        which is a dictionary based on the R object DiscountCurve. Column times is year fractions going forward in time.
        So today is 0, tomorrow is 1/365, a year from today is 1 etc...
    days_in_year : int
        360 or 365
    convention : list
        List of convention e.g. ["act",360], [30,360],...
    bus_calendar : str
        Banking day calendar. Not implemented.
    output : str
        "price" for swap price or "all" for price, cash flow data frame, duration.

    Returns
    -------
    returns present value as a scalar if output=='price. Otherwise returns dictionary with swap price, cash flow data frame and duration

    Examples
    --------
    >>> import risktools as rt
    >>> usSwapCurves = rt.data.open_data('usSwapCurves')
    >>> rt.swap_irs(
            trade_date="2020-01-04", 
            eff_date="2020-01-06", 
            mat_date="2022-01-06", 
            notional=1000000, 
            pay_rec = "rec", 
            fixed_rate=0.05, 
            float_curve=usSwapCurves, 
            reset_freq='Q', 
            disc_curve=usSwapCurves, 
            days_in_year=360, 
            convention="act", 
            bus_calendar="NY", 
            output = "all"
        )
    """
    if trade_date is None:
        trade_date = (_pd.Timestamp.now().floor("D"),)
    if eff_date is None:
        eff_date = trade_date + _pd.DateOffset(days=2)
    if mat_date is None:
        mat_date = eff_date + _pd.DateOffset(years=2)
    if float_curve is None:
        float_curve = us_swap
    if disc_curve is None:
        disc_curve = us_swap

    dates = custom_date_range(eff_date, mat_date, freq=reset_freq)

    # in case mat_date does not fall evenly on freq, take last date before
    dates = dates[dates <= mat_date]
    dates = _pd.Index([_pd.to_datetime(trade_date)]).append(dates)

    if (days_in_year in [360, 365]) == False:
        raise ValueError("days_in_year must be either 360 or 365")

    # placeholder
    if convention != "act":
        raise ValueError("function only defined for convention='act'")

    df = _pd.DataFrame(
        {
            "dates": dates,
            "day2next": (dates[1:] - dates[:-1]).days.append(
                _pd.Index([0])
            ),  # calc days to next period, short one element at end so add zero
            "times": (dates - dates[0]).days
            / 365,  # calc days to maturity from trade_date
        }
    )

    disc = _interpolate.splrep(disc_curve["times"], disc_curve["discounts"])
    df["disc"] = _interpolate.splev(df.times, disc)

    df["fixed"] = notional * fixed_rate * (df.day2next / days_in_year)
    df.loc[df.day2next <= 20, "fixed"] = 0
    df.fixed = df.fixed.shift() * df.disc

    disc_float = _interpolate.splrep(float_curve["times"], float_curve["discounts"])
    df["disc_float"] = _interpolate.splev(df.times, disc_float)
    df["floating"] = notional * (df.disc_float / df.disc_float.shift(-1) - 1)
    df.loc[df.day2next <= 20, "floating"] = 0
    df.floating = df.floating.shift() * df.disc
    df["net"] = df.fixed - df.floating

    df = df.fillna(0)
    pv = df.net.sum()
    df["duration"] = (
        (dates - _pd.to_datetime(trade_date)).days / days_in_year * df.net / pv
    )
    duration = df.duration.sum()

    if pay_rec == "pay":
        pv *= -1

    if output == "price":
        return pv
    else:
        return {"pv": pv, "df": df, "duration": duration}


def swap_com(df, futures_names, start_dt, end_dt, cmdty, exchange):
    """
    Function to calculate commodity swap pricing from futures contract exchange settlements of two contracts.
    Mostly used when the contract expires mid-month to work out the calendar month average

    Parameters
    ----------
    df : DataFrame
        Wide dataframe of futures prices
    futures_names : List[Tuple[str]]
        2 element List or tuple of futures contract names in order of expiry (i.e. [first expirying contract, second expirying contract])
    start_dt : str | datetime
        A string in the format of 'YYYY-mm-dd' or a datetime object
    end_dt : str | datetime
        A string in the format of 'YYYY-mm-dd' or a datetime object
    cmdty : str
        name of expiry contract type as specified in the expiry_table of this package. See risktools.data.open_data('expiry_table')['cmdty']
    exchange : ['nymex','ice']
        Name of exchange. Only 'nymex' and 'ice' supported. Used to filter the dataframe risktools.data.open_data('holidaysOil')

    Returns
    -------
    a dataframe of swap prices

    Examples
    --------
    >>> df = rt.get_prices(up['m*']['user'], up['m*']['pass'], codes=['CL0M','CL0N','CL0Q'], start_dt='2019-08-26')
    >>> df = df.settlement_price.unstack(level=0)
    >>> rt.swap_com(df=df, futures_names=["CL0M","CL0N"], start_dt="2020-05-01", end_dt="2020-05-30", cmdty="cmewti", exchange="nymex")
    """
    # get contract expiry dates
    exp = data.open_data("expiry_table")
    exp = exp.query(
        f"Last_Trade >= '{start_dt}' & Last_Trade <= '{end_dt}' & cmdty == '{cmdty}'"
    )["Last_Trade"]

    hol = data.open_data("holidaysOil").query(f"key == '{exchange}'")

    biz_dates = set(_pd.date_range(start_dt, end_dt, freq="B").to_list())
    biz_dates = biz_dates - set(hol.value)

    res = _pd.DataFrame(index=list(biz_dates)).sort_index()
    res["up2expiry"] = 1

    next_contract_dt = _pd.to_datetime(exp.values[0]) + _pd.DateOffset(days=1)

    res.loc[next_contract_dt:, "up2expiry"] = 0

    first_fut_weight = res.up2expiry.sum() / res.up2expiry.count()

    df = df[futures_names]
    df["swap"] = df.iloc[:, 0] * first_fut_weight + df.iloc[:, 1] * (
        1 - first_fut_weight
    )

    return df.dropna()


def get_ir_swap_curve(
    username, password, currency="USD", start_dt="2019-01-01", end_dt=None
):
    """
    Extract historical interest rate swap data for Quantlib DiscountsCurve function
    using Morningstar and FRED data

    Parameters
    ----------

    username : str
        Morningstar API username
    password : str
        Morningstar API password
    currency : str
        Not used - future enchancement
    start_dt : str | datetime, optional
        earliest date to return data from, by default "2019-01-01"
    end_dt : str | datetime, optional
        lastest date to return data from, by default None. If None, the function return everything from start_dt forward

    Examples
    --------
    >>> import risktools as rt
    >>> rt.get_ir_swap_curve(username=up['m*']['user'], password=up['m*']['pass'])
    """

    # fmt: off
    libor_ticks = _pd.DataFrame(dict(
        tick_nm = ["d1d","d1w","d1m","d3m","d6m","d1y"],
        type = ["ICE.LIBOR"]*6,
        source = ["FRED"]*6,
        codes = ["USDONTD156N","USD1WKD156N","USD1MTD156N","USD3MTD156N","USD6MTD156N","USD12MD156N"]
    ))

    mstar_ticks = _pd.DataFrame(dict(
        tick_nm = ["fut" + str(i) for i in range(1,9)],
        type = ["EuroDollar"]*8,
        source = ["Morningstar"]*8,
        codes = [f"ED_{str(i).zfill(3)}_Month" for i in range(1,9)]
    ))

    yrs = [2,3,5,7,10,15,20,30]
    irs_ticks = _pd.DataFrame(dict(
        tick_nm = [f"s{str(i)}y" for i in yrs],
        type = ["IRS"]*8,
        source = ["FRED"]*8,
        codes = [f"ICERATES1100USD{str(i)}Y" for i in yrs]
    ))

    # ticks = libor_ticks.append(irs_ticks).append(mstar_ticks)
    ticks = _pd.concat([libor_ticks, irs_ticks, mstar_ticks], axis=0)
    # fmt: on

    r = get_prices(
        codes=ticks.loc[ticks.source == "Morningstar", "codes"].to_list(),
        feed="CME_CmeFutures_EOD_continuous",
        username=username,
        password=password,
        start_dt=start_dt,
        end_dt=end_dt,
    )

    f = _pdr.DataReader(
        name=ticks.loc[ticks.source == "FRED", "codes"].to_list(),
        data_source="fred",
        start=start_dt,
        end=end_dt,
    )

    f.index.name = "Date"
    f.index = _pd.to_datetime(f.index)
    df = (
        r.unstack(0)
        .droplevel(0, 1)
        .merge(f, left_index=True, right_index=True, how="outer")
    )

    df[df.columns[df.columns.str.contains("ICERATES")]] /= 100
    df[df.columns[df.columns.str[:3] == "USD"]] /= 100
    df = df.rename(ticks.set_index("codes").tick_nm.to_dict(), axis=1)
    df = df[ticks.tick_nm]

    return df


def swap_info(
    username,
    password,
    date=None,
    tickers=["CL", "CL_001_Month"],
    feeds=[
        "Crb_Futures_Price_Volume_And_Open_Interest",
        "CME_NymexFutures_EOD_continuous",
    ],
    exchange="nymex",
    contract="cmewti",
    output="all",
):
    """
    Returns dataframe required to price a WTI averaging instrument based on first line settlements.

    Parameters
    ----------
    username : str
        Morningstar user name as character - sourced locally in examples.
    password : str
        Morningstar user password as character - sourced locally in examples.
    date : str | datetime
        Date as of which you want to extract daily settlement and forward values. By default None
    tickers : str | iterable[str]
        Morningstar tickers for get_curve() and get_prices() functions
    feeds : str | iterable[str]
        Feeds for Morningstar get_curve() and get_prices() functions. Must be in same 
        order as tickers parameter
    exchange : str
        Exchange code in holidaysOil from risktools.data.open_data function. Defaults to "nymex".
    contract : str
        Contract code in expiry_table.  from risktools.data.open_data function. See
        risktools.data.open_date("expiry_table").cmdty.unique() for options.
    output :str
        "chart" for chart, "dataframe" for dataframe or "all" dataframe and chart

    Returns
    -------
    Plot, dataframe or a dict of data frame and plot if output = "all".

    Examples
    --------
    >>> import risktools as rt
    >>> feeds = ["Crb_Futures_Price_Volume_And_Open_Interest",
            "CME_NymexFutures_EOD_continuous"]
    >>> tickers = ["CL","CL_001_Month"]
    >>> rt.swap_info(username, password, date="2020-05-06", tickers=tickers, feeds=feeds, contract="cmewti", exchange="nymex",
            output = "all")
    """
    # get today's futures curve prices by expiration date
    wti = get_curves(
        username=username,
        password=password,
        feed=feeds[0],
        contract_roots=tickers[0],
        fields=["Close"],
        date=date,
    )

    if date is None:
        date = _pd.Timestamp.now().floor("D")
    else:
        date = _pd.to_datetime(date)

    # broadcast wti expiration date to daily prices (effectively a stepped curve)
    drng = _pd.date_range(start=date, end=date + _pd.DateOffset(months=4), freq="B")

    cal = data.open_data("holidaysOil")
    cal = cal[cal.key == exchange]

    df = _pd.DataFrame(index=drng)
    # remove nymex holidays
    df = df[~df.index.isin(cal.value)]

    # combine dataframes together by broadcasting wti to wf on expiration date
    # and then backfill
    df.index.name = "date"
    wti = wti.rename({"expirationDate": "date"}, axis=1)
    df = df.join(wti.set_index("date")).bfill()

    hist = get_prices(
        username=username,
        password=password,
        feed=feeds[1],
        codes=tickers[1],
        start_dt=date + _pd.DateOffset(day=1),
    )

    hist = (
        hist.reset_index()[["Date", "settlement_price"]]
        .rename({"Date": "date", "settlement_price": "price"}, axis=1)
        .set_index("date")
    )
    hist["futures_contract"] = "first_line_settled"
    df = df.rename({"code": "futures_contract", "Close": "price"}, axis=1)[
        ["futures_contract", "price"]
    ]
    # df = (
    #     hist.loc[:(date), :]
    #     .append(df.loc[(date + _pd.DateOffset(days=1)) :, :])
    #     .dropna()
    # )

    df = (
        _pd.concat([
            hist.loc[:(date), :], 
            df.loc[(date + _pd.DateOffset(days=1)) :, :]
        ], 
        axis=0)
        .dropna()
    )

    if output == "dataframe":
        out = df
    else:
        title = (
            f"Swap Pricing: {contract}, Futures Curve as of {date.strftime('%Y-%m-%d')}"
        )

        fig = _px.scatter(
            df,
            y="price",
            color="futures_contract",
            title=title,
            labels={"price": "$/bbl"},
        )
        if output == "chart":
            out = fig
        else:
            out = dict(chart=fig, dataframe=df)

    return out


def swap_fut_weight(
    month, contract="cmewti", exchange="nymex", output="first_fut_weight"
):
    """
    Function used to calculate the calendar month average price of a futures curve. Some 
    futures curves such as NYMEX WTI expires mid-month (i.e. 2020-09-22) but some products
    are priced as differentials to the calendar month average (CMA) price of the contracts. 
    To calculate the CMA, up need the contract that's active for the first part of the month, 
    the contract that's active for the later half as well as the weighting to apply to both. 
    This function is designed to return the weights for contracts 1 and 2 as well as the weights

    Parameters
    ----------
    month : str | datetime
        First calendar day of the month in YYYY-MM-DD
    contract : str
        Exchange code in data.open_data("holidaysOil"). Currently only "nymex" and "ice" supported.
        By default "cmewti"
    exchange : str
        Contract code in data.open_data("expiry_table"). By default "nymex"
    output : str
        "num_days_fut1" to get the number of days from the beginning of the month up to and including the expiry date.
        "num_days_fut2" to get the number of days from the expiry date to month end.
        "first_fut_weight" the weight to assign to the futures contract for "month" as a ratio of the number of days before expiry (i.e. include in ). By default "first_fut_weight"

    Returns
    -------
    If first_fut_weight as dataframe, if not then an integer
    
    Examples
    --------
    >>> import risktools as rt
    >>> rt.swap_fut_weight(month="2020-09-01", contract="cmewti", exchange="nymex", output="first_fut_weight")
    """

    month = _pd.to_datetime(month) + _pd.DateOffset(day=1)
    month_end = month + _pd.DateOffset(months=1) - _pd.DateOffset(days=1)

    drng = _pd.date_range(month, month_end, freq="B")

    cal = data.open_data("holidaysOil")
    cal = cal[cal.key == exchange]

    df = _pd.DataFrame(index=drng)
    # remove nymex holidays
    df = df[~df.index.isin(cal.value)]

    exp = data.open_data("expiry_table")
    exp = exp[
        (exp["Last_Trade"] >= month)
        & (exp["Last_Trade"] <= month_end)
        & (exp["cmdty"] == contract)
    ].Last_Trade.values[0]

    df["up_to_expiry"] = _np.where(df.index <= exp, 1, 0)
    num_days_fut1 = df.up_to_expiry.sum()
    num_days_fut2 = df.shape[0] - df.up_to_expiry.sum()
    first_fut_weight = num_days_fut1 / df.shape[0]

    if output == "num_days_fut1":
        out = num_days_fut1
    elif output == "num_days_fut2":
        out = num_days_fut2
    else:
        out = first_fut_weight

    return out




import numpy as _np
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator as _AML


def _moment(x, k):
    m1 = x.mean()
    out = ((x - m1) ** k).sum() / len(x)
    return out


def _unbiased_skewness(x):
    sd = _np.sqrt(_moment(x, 2))
    n = len(x)
    gamma1 = _moment(x, 3) / sd ** 3
    skewness = _np.sqrt(n * (n - 1)) * gamma1 / (n - 2)
    return skewness


def _unbiased_kurtosis(x):
    n = len(x)
    var = _moment(x, 2)
    gamma2 = _moment(x, 4) / var ** 2
    kurtosis = (n - 1) / ((n - 2) * (n - 3)) * ((n + 1) * gamma2 - 3 * (n - 1)) + 3
    return kurtosis


def _unbiased_std(x):
    return x.std(ddof=1)


def _sample_skewness(x):
    sd = _np.sqrt(_moment(x, 2))
    skewness = _moment(x, 3) / sd ** 3
    return skewness


def _sample_kurtosis(x):
    var = _moment(x, 2)
    kurtosis = _moment(x, 4) / var ** 2
    return kurtosis


def _sample_std(x):
    return _np.sqrt(_moment(x, 2))


def _discrete_plot(ax, kurtmax):
    rmax = 100
    rmin = -rmax
    # negbin dist
    p = _np.exp(-10)
    lr = _np.arange(rmin, rmax + 0.1, 0.1)
    r = _np.exp(lr)
    s2a = (2 - p) ** 2 / (r * (1 - p))
    ya = 3 + 6 / r + p ** 2 / (r * (1 - p))
    p = 1 - _np.exp(-10)
    lr = _np.arange(rmax, rmin - 0.1, -0.1)
    r = _np.exp(lr)
    s2b = (2 - p) ** 2 / (r * (1 - p))
    yb = 3 + 6 / r + p ** 2 / (r * (1 - p))
    s2 = _np.append(s2a, s2b)
    y = _np.append(ya, yb)
    xy = _np.array([s2, y])
    ax.fill(xy[0, :], xy[1, :], "lightgrey", label="negative binomial")

    # if (!is.null(boot)) {
    # legend(xmax*0.2,ymax*0.98,pch=1,legend="bootstrapped values",
    #     bty="n",cex=0.8,col=boot.col)
    # }

    # poisson dist
    llambda = _np.arange(rmin, rmax + 0.1, 0.1)
    lamb = _np.exp(llambda)
    s2 = 1 / lamb
    y = 3 + 1 / lamb
    ax.plot(s2, y, "--", label="Poisson")
    ax.plot(0, 3, "*", label="normal")
    return ax


def _cont_plot(ax, kurtmax):
    rmax = 100
    rmin = -rmax
    # beta dist
    # fmt: off
    p=_np.exp(rmin)
    lq=_np.arange(rmin,rmax+0.1,0.1)
    q=_np.exp(lq)
    s2a=(4*(q-p)**2*(p+q+1))/((p+q+2)**2*p*q)
    ya=(3*(p+q+1)*(p*q*(p+q-6)+2*(p+q)**2)/(p*q*(p+q+2)*(p+q+3))) #kurtmax-
    p=_np.exp(rmax)
    lq=_np.arange(rmin,rmax+0.1,0.1)
    q=_np.exp(lq)
    s2b=(4*(q-p)**2*(p+q+1))/((p+q+2)**2*p*q)
    yb=(3*(p+q+1)*(p*q*(p+q-6)+2*(p+q)**2)/(p*q*(p+q+2)*(p+q+3))) #kurtmax-
    s2=_np.append(s2a,s2b)
    y=_np.append(ya,yb)
    xy = _np.array([s2,y])
    ax.fill(xy[0,:],xy[1,:], 'lightgrey', label='beta')
    # gamma dist
    lshape=_np.arange(rmin,rmax+0.1,0.1)
    shape=_np.exp(lshape)
    s2=4/shape
    y=(3+6/shape)
    ax.plot(s2,y,"--", label='gamma')
    # lnorm dist
    # I have no idea why this needs to be such a large
    # range? especially since we're then taking the exp
    # and the 4th power - quickly maxes out number length
    # override for now...
    # lshape=_np.arange(rmin,rmax+0.1,0.1)
    # shape=_np.exp(lshape)
    # es2=_np.exp(shape**2)
    es2=_np.arange(0,rmax+0.1,0.1)
    s2=(es2+2)**2*(es2-1)
    y=(es2**4+2*es2**3+3*es2**2-3)
    # added to only show +ve x once I overrode above
    y = y[s2>=0]
    s2 = s2[s2>=0]
   
    ax.plot(s2,y,":", label='lognormal')
    # fmt: on
    ax.plot(0, 3, "*", label="normal")
    ax.plot(0, 9 / 5, "^", label="uniform")
    ax.plot(2 ** 2, 9, "X", label="exponential")
    ax.plot(0, 4.2, "P", label="logistic")
    return ax


def _boot(x, boot, skewness, kurtosis):
    if boot < 10:
        raise ValueError("boot value is less than 10")

    n = len(x)

    data = _np.random.choice(x, (n, boot))
    skewboot = _np.apply_along_axis(skewness, 0, data)
    kurtboot = _np.apply_along_axis(kurtosis, 0, data)

    return skewboot, kurtboot


def describe_distribution(
    x, discrete=False, boot=None, method="unbiased", graph=True, ax=None, **plot_args
):
    """
    Function for generating population/sample statistics as well as plotting
    a Cullen and Frey graph for identifying distribution types based on data.

    Parameters
    ----------
    x : iterable[numeric]
        Data to be described
    discrete : bool
        True to show discrete distributions in the Cullen and Frey graph. By
        default False
    boot : int
        If not None, function will bootstrap samples from the data to generate
        bootstrapped statistics for the Cullen and Frey graph
    method : str["unbiased"|"sample"]
        Method to be used to generate data statistics. If "sample" is used, 
        statistics such as standard deviation, skewness and kurtosis will be
        calculated using the formulas for population samples (i.e. 1 degree of freedom).
        If "unbiased" is used, the calculations will use the formulas for the
        full population (i.e. 0 degrees of freedom). By default "unbiased".
    graph : bool 
        If True, the function will return the Cullen and Frey graph for the 
        data x. If False, it will return a dictionary with the population/sample
        statistics. By default True.
    ax : matplotlib Axes object
        Only valid if graph = True. A matplotib axes object can be passed so that
        any graph generated by this function to added to the passed Axes object.
        Useful for adding the Cullen and Frey graph to a subplot. By default None.
    **plot_args
        Parameters to be passed to the matplotlib.pyplot.subplots() function.
        For example figsize=(10,10)
    
    Returns
    -------
    graph = False : returns a dictionary
    graph = True & ax is None : returns a matplotlib Figure object
    graph = True & ax is not None : returns a matplotlib Axes object

    Examples
    --------
    >>> import risktools as rt
    >>> import matplotlib.pyplot as plt
    >>> x = [1,4,7,9,15,20,54]
    >>> rt.describe_distribution(x, method="sample", discrete=False, boot=500)
    >>> fig, ax = plt.subplots(1,2)
    >>> rt.describe_distribution(x, method="sample", discrete=False, boot=500, ax=ax[0], figsize=(10,10))
    """
    x = _np.array(x)

    if method == "unbiased":
        skewness = _unbiased_skewness
        kurtosis = _unbiased_kurtosis
        std = _unbiased_std
    else:
        skewness = _sample_skewness
        kurtosis = _sample_kurtosis
        std = _sample_std

    res = dict(
        min=x.min(),
        max=x.max(),
        median=_np.median(x),
        mean=x.mean(),
        sd=std(x),
        skewness=skewness(x),
        kurtosis=kurtosis(x),
        method=method,
    )

    skewdata = res["skewness"]
    kurtdata = res["kurtosis"]

    if graph:
        fig_flag = False
        if boot is not None:
            skewboot, kurtboot = _boot(x, boot, skewness, kurtosis)
            kurtmax = max(10, _np.ceil(kurtboot.max()))
            xmax = max(4, _np.ceil(skewboot.max() ** 2))
        else:
            kurtmax = max(10, _np.ceil(kurtdata))
            xmax = max(4, _np.ceil(skewdata ** 2))

        ymax = kurtmax  # in orginal code is kurtmax-1 but ugly graph with high kurt
        if ax is None:
            fig, ax = plt.subplots(**plot_args)
            fig_flag = True
        ax.set_ylim((ymax + 1, 0))
        ax.set_xlim((-0.1, xmax + 0.1))
        ax.set_title("Cullen and Frey Graph")
        ax.set_ylabel("kurtosis")
        ax.set_xlabel("square of skewness")
        ax.xaxis.set_minor_locator(_AML())
        ax.yaxis.set_minor_locator(_AML())

        # ax.xaxis.set_ticks(_np.arange(0, xmax + 1, 1))
        # ax.yaxis.set_ticks(_np.arange(0, ymax + 1, 1))

        if discrete == False:
            ax = _cont_plot(ax, kurtmax)
        else:
            ax = _discrete_plot(ax, kurtmax)

        # bootstrap sample for observed distribution
        if boot is not None:
            ax.plot(
                skewboot ** 2, kurtboot, ".", alpha=0.2, label="bootstrapped values"
            )

        ax.plot([skewdata ** 2], [kurtdata], "bo", label="observation")
        ax.legend()
    if graph == True:
        if fig_flag == True:
            return fig
        else:
            return ax
    else:
        return res


if __name__ == "__main__":
    x = _np.array([1, 4, 7, 9, 15, 20, 54])
    n = len(x)

    describe_distribution(x, method="sample")




# multivariate simulations

from ast import arguments
import numpy as _np
import pandas as _pd
import matplotlib.pyplot as _plt
import plotly.graph_objects as _go
from ._sims import fitOU, simOU, simOUJ
from abc import ABC as _ABC, abstractmethod as _abstractmethod
from numpy.random import Generator, SFC64


def calc_spread_MV(df, formulas):
    """
    Calculate a series of spreads for a multivariate stochastic process.

    Parameters
    ----------
    df : DataFrame
        DataFrame containing the simulated values of the stochastic processes.
        The columns correspond to the assets and the index corresponds to the
        time steps.
    formulas : dictionary
        Dictionary of formulas to use for calculating the spreads. The key must be
        the name of the spread and the value must be a string containing the formula
        for calculating the spread. The formula must be a valid Python expression using
        the names of the columns in df as variables. For example, to calculate the spread
        between asset_1 and asset_2 less 5, the formula would be 'asset_1 - asset_2 - 5'.

    Returns
    -------
    DataFrame containing the simulated values of the spreads. The columns
    correspond to the spreads and the index corresponds to the time steps.

    Example
    -------
    >>> import risktools as rt
    >>> df = rt.simGBM_MV([100, 100], 0.05, [0.2, 0.3], 1, 0.01, cor=[[1, 0.5], [0.5, 1]], sims=10)
    >>> rt.calc_spread_MV(df, {'spread':'1-2'})
    """
    spreads = _pd.DataFrame(index=df.index)

    for i, r in enumerate(formulas.keys()):
        spreads[i] = df.eval(formulas[r])

    spreads.columns = list(formulas.keys())

    return spreads


def fitOU_MV(df, dt, log_price=False, method="OLS", verbose=False):
    """
    Fit multiple OU processes

    Parameters
    ----------
    df : DataFrame
        DataFrame of multiple OU processes where each process is a column.  Thus
        a matrix of size (p x N) where p is the number of time steps and N is
        the number of assets.
    dt : float
        Assumed time step for the OU processes. Must be the same for all
        OU processes.
    log_price : bool
        If True, the spread is assumed to be log prices and the log of the spread is taken.
        All price series must be log_prices or all not. Default is False.
    method : ['OLS' | 'MLE'], optional
        Method used to fit OU process. OLE for ordinary least squares and MLE for
        Maximum Liklihood. By default 'OLS'.
    verbose : bool
        If True, prints the estimated parameters. Only used is method is OLS.
        Default is False.

    Returns
    -------
    DataFrame of the fitted parameters for each OU process. The rows
    correspond to the parameters and the columns correspond to the OU processes.

    Example
    -------
    >>> import risktools as rt
    """

    params = _pd.DataFrame()

    for c in df.columns:
        ret = fitOU(df[c], dt, log_price=log_price, method=method, verbose=verbose)
        params.loc["theta", c] = ret["theta"]
        params.loc["annualized_sigma", c] = ret["annualized_sigma"]
        params.loc["mu", c] = ret["mu"]

    return params


def generate_eps_MV(cor, T, dt, sims=1000, mu=None, seed=None):
    """
    Generate epsilons from a multivariate normal distribution
    for use in multivariate stochastic simulations

    Parameters
    ----------
    cor : matrix-like[float]
        Correlation matrix of the OU processes. Must be a square matrix of
        size M x M and positive definite where M is the number of stochastic
        processes.
    T : float
        Time horizon of the simulation (in years).
    dt : float
        Time step of the simulation (in years).
    sims : int
        Number of simulations. By default 1000.
    mu : array-like[float], optional
        Array of means to use for the multivariate normal for each random process.
        If None, mu = 0 is used for all random processes. By default None.
    seed : int
        To pass to numpy random number generator as seed. For testing only.

    Returns
    -------
    Matrix of random numbers to use for the simulation. The first dimension
    corresponds to the time steps, the second dimension corresponds to the simulations,
    and the third dimension corresponds to the OU processes.

    Example
    -------
    >>> import risktools as rt
    >>> rt.generate_eps_MV([0.2, 0.3], [[1, 0.5], [0.5, 1]], 1, 0.01, 10)
    """
    N = int(T / dt)

    # if ~isinstance(sigma, _np.ndarray):
    #     sigma = _np.array(sigma)
    if ~isinstance(cor, _np.ndarray):
        cor = _np.array(cor)

    if mu is not None:
        if ~isinstance(mu, _np.ndarray):
            mu = _np.array(mu)
    else:
        mu = _np.zeros(cor.shape[0])

    # unneeded since we multiple by sigma in simOU/GBM
    # sd = _np.diag(sigma)
    # cov = sd @ cor @ sd
    cov = cor
    rng = Generator(SFC64(seed))

    eps = rng.multivariate_normal(mu, cov, size=(N, sims))

    # eps = _np.random.multivariate_normal(mu, cov, size=(N, sims))

    return eps


def simGBM_MV(s0, r, sigma, T, dt, mu=None, cor=None, eps=None, sims=1000):
    """
    Simulate Geometric Brownian Motion for stochastic processes with
    multiple assets using a multivariate normal distribution.

    Parameters
    ----------
    s0 : array-like
        Initial values of the stochastic processes. Must be a 1D array of length
        M where M is the number of assets.
    r : float
        Risk-free rate.
    sigma : array-like
        Volatility of the stochastic processes (annualized standard deviations of
        returns). Must be a 1D array of length M where M is the number of assets.
        Only used if eps is None.
    T : float
        Time horizon of the simulation (in years).
    dt : float
        Time step of the simulation (in years).
    mu : array-like, optional
        Means to use for multivariate normal distribution of returns. Must be
        a 1D array of length M where M is the number of assets. If None, mu = 0
        is used for all assets. Only used if eps is None.
    cor : matrix-like
        Correlation matrix of the stochastic processes. Must be a square matrix of
        size M x M and positive definite where M is the number of assets.
        Only used if eps is None.
    eps : array-like, optional
        Random numbers to use for the simulation. If not provided, random numbers are
        generated using a multivariate normal distribution. Must be a 3-D array of
        size (N x sims x M) where N is the number of time steps, sims is the number of
        simulations, and M is the number of assets. By default None.
    sims : int
        Number of simulations. By default 1000.

    Returns
    -------
    Matrix of simulated values of the stochastic processes. The first dimension
    corresponds to the time steps, the second dimension corresponds to the simulations,
    and the third dimension corresponds to the assets.

    Example
    -------
    >>> import risktools as rt
    >>> rt.simGBM_MV(s0=[100,100], r=0.0, sigma=[0.1,0.1], T=1, dt=1/252, cor=[[1,0],[0,1]], sims=100)
    """
    if (cor is None) & (eps is None):
        raise ValueError("correlation matrix cor required if eps not passed")

    if ~isinstance(s0, _np.ndarray):
        s0 = _np.array(s0)
    if ~isinstance(sigma, _np.ndarray):
        sigma = _np.array(sigma)
    if ~isinstance(r, _np.ndarray):
        r = _np.array(r)
    if ~isinstance(cor, _np.matrix):
        cor = _np.matrix(cor)

    if mu is not None:
        if ~isinstance(mu, _np.ndarray):
            mu = _np.array(mu)
    else:
        mu = _np.zeros(len(s0))

    N = int(T / dt)

    if eps is None:
        eps = generate_eps_MV(cor, T, dt, sims, mu)

    s = _np.zeros((N + 1, sims, len(s0)))

    for i in range(0, s.shape[2]):
        s[1:, :, i] = _np.exp(
            (r - 0.5 * sigma[i] ** 2) * dt + sigma[i] * _np.sqrt(dt) * eps[:, :, i]
        )
    s[0, :, :] = s0

    return s.cumprod(axis=0)


def simOU_MV(
    s0,
    mu,
    theta,
    sigma,
    T,
    dt=None,
    cor=None,
    sims=1000,
    eps=None,
    seed=None,
    log_price=False,
    **kwargs
):
    """
    Simulate Ornstein-Uhlenbeck process for stochastic processes for
    multiple assets using a multivariate normal distribution.

    Parameters
    ----------
    s0 : array-like[float]
        Initial values of the stochastic processes. Must be a 1D array of length
        M where M is the number of assets.
    mu : array-like[float]
        Mean of the OU processes. Must be a 1D or 2D array.

        1D Arrays:
            If a 1D array is provided, each element of the array represents the mean of each asset.
            Array must of size M where M is the number of assets.
        2D Arrays:
            A 2D array can be provided to varying the mean of each asset over time. The size of the
            array must be N x M where N is the number of time steps and M is the number of assets.
    theta : array-like[float]
        Mean reversion parameter of the OU processes. Must be a 1D array of length M
        where M is the number of assets.
    sigma : array-like[float]
        Annualized volatility or standard deviation. To calculate, take daily volatility and multiply by sqrt(T/dt).
        Can be either a 1D, 2D, or 3D array:

        1D Arrays:
            If a 1D array is provided, each element of the array represents the volatility of each asset,
            is used for all time steps, all simulations. 1D arrays must of size M where M is the number of assets.
        2D Arrays:
            A 2D array can be provided to varying the volatility of each asset over time. The size of the
            array must be N x M where N is the number of time steps and M is the number of assets.
        3D Arrays:
            A 3D array can be provided to varying the volatility of each asset over time and over each
            simulation (stochastic volatility). The size of the array must be N x sims x M where N is the number of time
            steps, sims is the number of simulations, and M is the number of assets.
    T : float
        Time horizon of the simulation (in years).
    dt : float, optional
        Time step of the simulation (in years). Not used if eps is provided. By default None.
    cor : matrix-like[float], optional
        Correlation matrix of the OU processes. Must be a square matrix of
        size M x M and positive definite. Only used if eps is None.
    sims : int, optional
        Number of simulations. By default 1000. Not used if eps is provided.
    eps : matrix-like, optional
        Random numbers to use for the simulation. Must be a 3D array of
        size (N x sims x M) where N is the number of time steps, sims is the number of
        simulations, and M is the number of assets. By default None.
    seed : int, optional
        To pass to numpy random number generator as seed. For testing only.
    log_price : bool
        Adds adjustment term to the mean reversion term if the prices passed are log prices. By
        default False.
    **kwargs : optional
        Keyword arguments to pass to simOU function.

    Returns
    -------
    Matrix of simulated values of the stochastic processes of size N x sims x M where
    N corresponds to the number of time steps, sims corresponds to the simulations,
    and M corresponds to the number of assets.

    Example
    -------
    >>> import risktools as rt
    >>> rt.simOU_MV(s0=[100,100], mu=[0.1,0.1], sigma=[0.3,0.3], theta=[10,10],
            T=1, dt=1/252, cor=[[1,0],[0,1]], sims=100)
    """
    if eps is None:
        if dt is None:
            raise ValueError("Must provide dt if eps is not provided.")
        if cor is None:
            raise ValueError("Must provide cor if eps is not provided.")
        eps = generate_eps_MV(cor=cor, T=T, dt=dt, sims=sims, seed=seed)
    else:
        dt = T / eps.shape[0]

    sigma_arr = _np.zeros(eps.shape)
    mu_arr = _np.zeros((eps.shape[0], eps.shape[2]))
    sigma = _np.array(sigma)
    mu = _np.array(mu)

    # broadcast sigma into 3D array
    if len(sigma.shape) == 1:

        for i in range(0, eps.shape[2]):
            sigma_arr[:, :, i] = sigma[i]
    elif len(sigma.shape) == 2:
        for i in range(0, eps.shape[2]):
            sigma_arr[:, :, i] = sigma[:, _np.newaxis, i]
    elif len(sigma.shape) == 3:
        sigma_arr = sigma

    # broadcast mu into 2D array
    if len(mu.shape) == 1:
        for i in range(0, eps.shape[2]):
            mu_arr[:, i] = mu[i]
    else:
        mu_arr = mu

    # number of time steps
    N = eps.shape[0]

    s = _np.zeros((N + 1, eps.shape[1], eps.shape[2]))

    for i in range(0, eps.shape[2]):
        s[:, :, i] = simOU(
            s0=s0[i],
            mu=mu_arr[:, i],
            theta=theta[i],
            sigma=sigma_arr[:, :, i],
            T=T,
            dt=dt,
            eps=eps[:, :, i],
            log_price=log_price,
            **kwargs
        )

    return s


def simOUJ_MV(
    s0,
    mu,
    theta,
    sigma,
    jump_prob,
    jump_avgsize,
    jump_stdv,
    T,
    dt=None,
    cor=None,
    sims=1000,
    mr_lag=None,
    eps=None,
    elp=None,
    ejp=None,
    seed=None,
    **kwargs
):
    """
    Simulate Ornstein-Uhlenbeck Jump process for stochastic processes for
    multiple assets using a multivariate normal distribution.

    Parameters
    ----------
    s0 : array-like[float]
        Initial values of the stochastic processes. Must be a 1D array of length
        M where M is the number of assets.
    mu : array-like[float]
        Mean of the OU processes. Must be a 1D or 2D array.

        1D Arrays:
            If a 1D array is provided, each element of the array represents the mean of each asset.
            Array must of size M where M is the number of assets.
        2D Arrays:
            A 2D array can be provided to varying the mean of each asset over time. The size of the
            array must be N x M where N is the number of time steps and M is the number of assets.
    theta : array-like[float]
        Mean reversion parameter of the OU processes. Must be a 1D array of length M
        where M is the number of assets.
    sigma : array-like[float]
        Annualized volatility or standard deviation. To calculate, take daily volatility and multiply by sqrt(T/dt).
        Can be either a 1D, 2D, or 3D array:

        1D Arrays:
            If a 1D array is provided, each element of the array represents the volatility of each asset,
            is used for all time steps, all simulations. 1D arrays must of size M where M is the number of assets.
        2D Arrays:
            A 2D array can be provided to varying the volatility of each asset over time. The size of the
            array must be N x M where N is the number of time steps and M is the number of assets.
        3D Arrays:
            A 3D array can be provided to varying the volatility of each asset over time and over each
            simulation (stochastic volatility). The size of the array must be N x sims x M where N is the number of time
            steps, sims is the number of simulations, and M is the number of assets.
    jump_prob : array_like[float]
        Probablity of jumps for a Possion process. Must be a 1D array of length M
        where M is the number of assets.
    jump_avgsize : array_like[float]
        Average size of jumps for a log normal distribution. Must be a 1D array of length M
        where M is the number of assets.
    jump_stdv : array_like[float]
        Standard deviation of average jump size for a log normal distribution. Must be a 1D array of length M
        where M is the number of assets.
    T : float
        Time horizon of the simulation (in years).
    dt : float, optional
        Time step of the simulation (in years). Not used if eps is provided. By default None.
    cor : matrix-like[float], optional
        Correlation matrix of the OU processes. Must be a square matrix of
        size M x M and positive definite. Only used if eps is None.
    sims : int, optional
        Number of simulations. By default 1000. Not used if eps is provided.
    mr_lag : array_like[int]
        Lag in mean reversion. If None, then no lag is used. If > 0, then the diffusion does not immediately
        return the mean after a jump at theta but instead with remain near the jump level for mr_lag periods.
        By default, this is None. Must be a 1D array of length M where M is the number of assets.
    eps : matrix-like, optional
        Random numbers to use for the simulation. Must be a 3D array of
        size (N x sims x M) where N is the number of time steps, sims is the number of
        simulations, and M is the number of assets. By default None. If None, then random numbers are generated.
    elp : numpy array, optional
        Array of random numbers to use for the log price jump.
        Must be a 3D array of
        size (N x sims x M) where N is the number of time steps, sims is the number of
        simulations, and M is the number of assets. By default None. If None, then random numbers are generated.
    ejp : numpy array, optional
        Array of random numbers to use for the jump size.
        Must be a 3D array of
        size (N x sims x M) where N is the number of time steps, sims is the number of
        simulations, and M is the number of assets. By default None. If None, then random numbers are generated.
    seed : int, optional
        To pass to numpy random number generator as seed. For testing only.
    **kwargs : optional
        Keyword arguments to pass to simOUJ function.

    Returns
    -------
    Matrix of simulated values of the stochastic processes of size N x sims x M where
    N corresponds to the number of time steps, sims corresponds to the simulations,
    and M corresponds to the number of assets.

    Example
    -------
    >>> import risktools as rt
    >>> rt.simOUJ_MV(s0=[100,100], mu=[0.1,0.1], sigma=[0.3,0.3], theta=[10,10],
            jump_prob=[0.1,0.1], jump_avgsize=[10,10], jump_stdv=[0.1,0.1],
            T=1, dt=1/252, cor=[[1,0],[0,1]], sims=100)
    """
    if eps is None:
        if dt is None:
            raise ValueError("Must provide dt if eps is not provided.")
        if cor is None:
            raise ValueError("Must provide cor if eps is not provided.")
        eps = generate_eps_MV(cor=cor, T=T, dt=dt, sims=sims, seed=seed)
    else:
        dt = T / eps.shape[0]

    sigma_arr = _np.zeros(eps.shape)
    mu_arr = _np.zeros((eps.shape[0], eps.shape[2]))
    sigma = _np.array(sigma)
    mu = _np.array(mu)

    # broadcast sigma into 3D array
    if len(sigma.shape) == 1:

        for i in range(0, eps.shape[2]):
            sigma_arr[:, :, i] = sigma[i]
    elif len(sigma.shape) == 2:
        for i in range(0, eps.shape[2]):
            sigma_arr[:, :, i] = sigma[:, _np.newaxis, i]
    elif len(sigma.shape) == 3:
        sigma_arr = sigma

    # broadcast mu into 2D array
    if len(mu.shape) == 1:
        for i in range(0, eps.shape[2]):
            mu_arr[:, i] = mu[i]
    else:
        mu_arr = mu

    # number of time steps
    N = eps.shape[0]

    if mr_lag is None:
        mr_lag = [mr_lag] * eps.shape[2]

    s = _np.zeros((N + 1, eps.shape[1], eps.shape[2]))

    for i in range(0, eps.shape[2]):
        ejp_tmp = ejp[:, :, i] if ejp is not None else None
        elp_tmp = elp[:, :, i] if elp is not None else None

        s[:, :, i] = simOUJ(
            s0=s0[i],
            mu=mu_arr[:, i],
            theta=theta[i],
            sigma=sigma_arr[:, :, i],
            jump_prob=jump_prob[i],
            jump_avgsize=jump_avgsize[i],
            jump_stdv=jump_stdv[i],
            T=T,
            dt=dt,
            eps=eps[:, :, i],
            elp=elp_tmp,
            ejp=ejp_tmp,
            seed=seed,
            mr_lag=mr_lag[i],
            **kwargs
        )

    return s


def generate_random_portfolio_weights(number_assets, number_sims=2500):
    """
    Generate a matrix of random portfolio weights based on
    a number of assets using a uniform distribution.

    Parameters
    ----------
    number_assets : int
        Number of assets in the portfolio.
    number_sims : int, optional
        Number of simulations. By default 2500.

    Returns
    -------
    2-D matrix of random portfolio weights. The first dimension corresponds to the
    simulations and the second dimension corresponds to the assets.

    Example
    -------
    >>> import risktools as rt
    >>> rt.generate_random_portfolio_weights(5, 1000)
    """
    weights = _np.random.uniform(size=(number_sims, number_assets))
    weights = _np.multiply(weights.T, 1 / weights.sum(axis=1)).T

    return weights


def calculate_payoffs(df, strike=0, payoff_funcs=None):
    """
    Calculate the payoffs for a series of simulated assets using asset specific payoff functions.

    Parameters
    ----------

    df : array-like[float]
        Array of (m x n x N) floats where m is the number of periods that the assets are simulated
        forward in time, n is the number of simulations run and N is the number of assets.
    strike : float, optional
        Only used if payoff_funcs is None. Strike price of the option. By default None.
    payoff_funcs : array-like[function], optional
        Array of payoff functions. Must be a 1D array of length N. Each function must take a
        single argument (the simulated asset price) and return a single value (the payoff) along
        the time axis. By default None. If none, the payoff is the max of the asset price and strike price at
        time T.

    Returns
    -------
    Matrix of payoffs for each simulation. The first dimension corresponds to the
    simulations and the second dimension corresponds to the assets.

    Example
    -------
    >>> import risktools as rt
    >>> def payoff(x):
            ret = _np.clip(x, 0, None)
            return ret.sum(axis=0)
    >>> rt.calc_payoffs(df, payoff_funcs=[payoff, payoff])
    """
    if payoff_funcs is None:
        payoffs = _np.clip(df - strike, 0, None)[-1, :, :]
    else:
        if len(payoff_funcs) != df.shape[2]:
            raise ValueError("Must provide a payoff function for each asset.")

        payoffs = _np.zeros((df.shape[1], df.shape[2]))

        for i in range(0, len(payoff_funcs)):
            payoffs[:, i] = payoff_funcs[i](df[:, :, i])

    return payoffs


def simulate_efficient_frontier(assets, weights):
    """
    Generate portfolio expected returns and risk using simulated
    asset prices and randomized weights.

    Parameters
    ----------
    assets : array-like[float]
        Array of (m x N) floats where m is the simulations of the assets and N is the number of assets.
    weights : array-like[float]
        Array of (n x N) floats where n is the number of portfolios and N is the number of assets.

    Returns
    -------
    Matrix of simulated portfolios. The first dimension corresponds to the
    simulations and the second dimension corresponds to the assets.

    Example
    -------
    >>> import risktools as rt
    >>> assets = rt.simGBM_MV(s0=[100,100], r=0.01, sigma=[0.1,0.1], T=1, dt=1/252, cor=[[1,0.5],[0.5,1]])
    >>> assets = rt.calc_payoffs(assets)
    >>> weights = rt.generate_random_portfolio_weights(5, 1000)
    >>> rt.sim_efficient_frontier(assets, weights)
    """

    out = _np.zeros((weights.shape[0], 2))

    for i in range(0, weights.shape[0]):
        ret = assets @ weights[i]

        out[i, 0] = ret.std()  # risk
        out[i, 1] = ret.mean()  # return

    return out


def make_efficient_frontier_table(returns, weights, asset_names=None):
    """
    Produce dataframe with portfolio returns, risk, and asset weights.

    Parameters
    ----------
    returns : array-like[float]
        Numpy array of (n x 2) floats where n in the number of portfolios simulated. First
        columns is the standard deviation or risk associated with the portfolio and the second
        column is the expected return.
    weights : array-like[float]
        Numpy array of (n x N) floats where n in the number of simulated portfolios and N in the
        number of assets in the portfolio.
    asset_names : list[str]
        List of strings to use as asset names in the output dataframe. Should be of length N
        where N is the number of assets in the portfolio.

    Returns
    -------
    Dataframe with the portfolio risk, expected return and asset weights as columns

    Examples
    --------
    >>> import risktools as rt
    >>> assets = rt.simGBM_MV(s0=[100,100], r=0.01, sigma=[0.1,0.1], T=1, dt=1/252, cor=[[1,0.5],[0.5,1]])
    >>> assets = rt.calc_payoffs(assets)
    >>> weights = rt.generate_random_portfolio_weights(5, 1000)
    >>> port = rt.sim_efficient_frontier(assets, weights)
    >>> port = rt.make_efficient_frontier_table(port, weights, asset_names=['A', 'B'])
    """
    out = _pd.DataFrame(_np.zeros((weights.shape[0], weights.shape[1] + 2)))

    out.iloc[:, [0, 1]] = returns
    out.iloc[:, 2:] = weights

    if asset_names is None:
        asset_names = list(range(1, weights.shape[1] + 1))

    out.columns = ["Risk", "Expected Return"] + asset_names

    return out


def plot_efficient_frontier(df):
    """
    Plot the efficient frontier for simulated portfolios stored in df.

    Parameters
    ----------
    df : DataFrame
        Must be a dataframe with the first two columns being the risk and expected return of the
        portfolio and the remaining columns being the weights of the assets in the portfolio.

    Returns
    -------
    Plot of the efficient frontier.

    Examples
    --------
    >>> import risktools as rt
    >>> assets = rt.simGBM_MV(s0=[100,100], r=0.01, sigma=[0.1,0.1], T=1, dt=1/252, cor=[[1,0.5],[0.5,1]])
    >>> assets = rt.calc_payoffs(assets)
    >>> weights = rt.generate_random_portfolio_weights(5, 1000)
    >>> port = rt.sim_efficient_frontier(assets, weights)
    >>> port = rt.make_efficient_frontier_table(port, weights, asset_names=['A', 'B'])
    >>> rt.plot_efficient_frontier(port)
    """

    df = df.copy()
    N = df.shape[1] - 2

    # convert weights columns into a list of strings with the asset names
    # fmt: off
    for i in range(0, N):
        df.iloc[:, i + 2] = (
            list(df.columns.astype(str))[i + 2] + " = " + df.iloc[:, i + 2].round(2).astype(str)
        )
    # fmt: on

    # build labels for portfolio weights
    df["text"] = df.iloc[:, 2:].apply(
        lambda x: "[" + ", ".join(x.dropna().astype(str)) + "]", axis=1
    )

    fig = _go.Figure()

    fig.add_trace(
        _go.Scattergl(
            x=df.iloc[:, 0],
            y=df.iloc[:, 1],
            marker_color=df.iloc[:, 1] / df.iloc[:, 0],
            mode="markers",
            marker_colorscale="Viridis",
            text=df["text"],
            showlegend=False,
            marker_colorbar=dict(
                title="Sharpe Ratio",
                titleside="right",
                tickmode="array",
            ),
        )
    )

    fig.update_layout(
        title="Efficient Frontier",
        title_x=0.5,
        title_xanchor="center",
        xaxis_title="Risk",
        yaxis_title="Expected Returns",
    )

    fig.update_traces(hovertemplate=None)

    return fig


def plot_portfolio(df, weights, fig, weight_names=None, label=True):
    """
    Plot a single portfolio on an efficient frontier chart

    Parameters
    ----------
    df : DataFrame[float] | array-like[float]
        Dataframe or array of simulated asset payoffs from the calc_payoffs function.
    weights : array-like[float]
        Array of asset weights for the portfolio to be plotted. Must sum to 1.
    fig : Figure
        Figure object to plot the portfolio on.
    weight_names : list[str]
        List of strings to use as asset names in the output dataframe. Should be of length N
        where N is the number of assets in the portfolio.
    label : bool
        If True, the portfolio marker will be labeled with the asset weights.

    Returns
    -------
    Plot of the efficient frontier with the portfolio added.

    Examples
    --------
    >>> import risktools as rt
    >>> assets = rt.simGBM_MV(s0=[100,100], r=0.01, sigma=[0.1,0.1], T=1, dt=1/252, cor=[[1,0.5],[0.5,1]])
    >>> assets = rt.calc_payoffs(assets)
    >>> weights = rt.generate_random_portfolio_weights(5, 1000)
    >>> port = rt.sim_efficient_frontier(assets, weights)
    >>> port = rt.make_efficient_frontier_table(port, weights, asset_names=['A', 'B'])
    >>> fig = rt.plot_efficient_frontier(port)
    >>> rt.plot_portfolio(assets, [.5, .5], fig)
    """

    if isinstance(df, _pd.DataFrame):
        df = df.to_numpy()

    # calculate portfolio
    df = df @ _np.array(weights)

    # construct portfolio labels
    if weight_names is not None:
        tmp = zip(weight_names, weights)
        weights = [str(j[0]) + " = " + str(round(j[1], 2)) for j in tmp]
    else:
        weights = [str(round(j, 2)) for j in weights]

    text = "[" + ", ".join([str(w) for w in weights]) + "]"

    x = [df.std()]
    y = [df.mean()]

    # add portfolio to efficient frontier chart
    fig.add_trace(
        _go.Scattergl(
            x=x,
            y=y,
            text=text,
            marker_color="red",
            mode="markers",
            legendgroup="hide",
            showlegend=False,
            marker_size=15,
        )
    )

    if label == True:
        fig.add_annotation(
            x=x[0],
            y=y[0],
            text=text,
            showarrow=False,
            yshift=20,
            font_size=12,
            font_color="red",
        )

    return fig


def shift(xs, n=1):
    # shift along first axis only
    e = _np.empty_like(xs)
    if n >= 0:
        e[:n] = _np.nan
        e[n:] = xs[:-n]
    else:
        e[n:] = _np.nan
        e[:n] = xs[-n:]
    return e


class _MVSIM(_ABC):
    """
    Abstract base class for multivariate simulation classes for
    calculating the payoffs of a portfolio of assets.
    """

    _frontier = None
    _sims = None
    _prices = None
    _params = None
    _asset_names = None

    @_abstractmethod
    def fit():
        pass

    @_abstractmethod
    def simulate():
        pass

    def plot_efficient_frontier(self, strike=0, payoff_funcs=None, portfolio_sims=5000):
        """
        Plot the efficient frontier based on a specificied payoff function.

        Parameters
        ----------
        strike : float, optional
            Only used if payoff_funcs is None. Strike price of the option. By default 0.
        payoff_funcs : array-like[function], optional
            Array of payoff functions. Must be a 1D array of length N. Each function must take a
            single argument (the simulated asset price) and return a single value (the payoff) along
            the time axis. By default None. If none, the payoff is the max of the asset price and strike price at
            time T.
        portfolio_sims : int
            Number of random portfolios to simulate for the efficient frontier.

        Returns
        -------
        Plot of the efficient frontier.

        Examples
        --------
        >>> import risktools as rt

        """

        self._payoffs = calculate_payoffs(self._sims, strike, payoff_funcs)

        # calculate efficient frontier
        weights = generate_random_portfolio_weights(len(self._s0), portfolio_sims)
        port = simulate_efficient_frontier(self._payoffs, weights)

        # make dataframe
        port = make_efficient_frontier_table(
            port, weights, asset_names=self._asset_names
        )

        # plot efficient frontier
        self._frontier = plot_efficient_frontier(port)

        return self._frontier

    def plot_portfolio(self, weights, label=True, weight_names=True):
        """
        Plot a single portfolio on an efficient frontier chart

        Parameters
        ----------
        df : DataFrame[float] | array-like[float]
            Dataframe or array of simulated asset payoffs from the calc_payoffs function.
        weights : array-like[float]
            Array of asset weights for the portfolio to be plotted. Must sum to 1.
        fig : Figure
            Figure object to plot the portfolio on.
        label : bool
            If True, the portfolio marker will be labeled with the asset weights.
        weight_names : bool
            Include weight names in label if already defined in object

        Returns
        -------
        Plot of the efficient frontier with the portfolio added.
        """

        if self._frontier is None:
            ValueError(
                "plot_efficient_frontier method must be run prior to this method"
            )

        names = self._asset_names if weight_names else None

        self._frontier = plot_portfolio(
            self._payoffs, weights, self._frontier, names, label
        )

        return self._frontier

    @property
    def sims(self):
        """
        Returns a 3-D array of size (p x sims x N) where p is the number of
        time steps, sims is the number of simulations, and N is the number
        of assets
        """
        return self._sims

    @property
    def prices(self):
        if self._prices is None:
            return "No time series we given to initialize the objects"
        else:
            return self._prices

    @property
    def parameters(self):
        params = self._params.copy()
        params.loc["s0", :] = self._s0
        return params

    def plot_sim(self, asset_num, sims=100, **kwargs):
        """
        Function to plot a subset of the simulates run for the asset
        specified by asset_num.

        Parameters
        ----------
        asset_num : int
            Column number of the asset to be shown from 0 to N-1 where
            N is the number of assets passed to the object.
        sims : int
            The number of simulations to show in the chart. Recommend
            around 100. By default 100.
        kwargs
            parameters to be passed to the pandas dataframe plot method.

        Returns
        -------
        matplotlib figure object.
        """

        return (
            _pd.DataFrame(self._sims[:, :, asset_num])
            .iloc[:, :sims]
            .plot(legend=False, **kwargs)
        )


class MVGBM(_MVSIM):
    """
    Class for simulating a multivariate GBM process for a portfolio of assets

    Parameters
    ----------
    r : array-like[float]
        Risk free rate of each asset in the portfolio.
    T : float
        Time to maturity of the portfolio.
    dt : float
        Time step for the simulation.
    s0 : array-like[float], optional
        Initial price of each asset in the portfolio.
        Required if prices not passed.
    sigma : array-like[float], optional
        Volatility of each asset in the portfolio.
        Required if prices not passed.
    cor : array-like[float], optional
        Correlation matrix for the assets in the portfolio.
        Required if prices not passed.
    prices : array-like[float], optional
        Matrix or dataframe of asset prices where the columns are
        different assets. Used to calibrate model using fit
        method if s0, sigma and cor not passed.
    asset_names : list[str], optional
        List of strings to use as asset names in the output dataframe.

    Example
    -------
    >>> import risktools as rt
    >>> mvgbm = rt.MVGBM(
            s0=[100,100, 100],
            r=0.01,
            sigma=[0.1,0.1,0.1],
            T=1,
            dt=1/252,
            cor=[[1,0.5,0.5],[0.5,1,0.5],[0.5,0.5,1]],
            asset_names=['A','B','C']
        )
    >>> mvgbm.fit()
    >>> mvgbm.simulate()
    >>> mvgbm.plot_efficient_frontier()
    """

    def __init__(
        self, r, T, dt, s0=None, sigma=None, cor=None, prices=None, asset_names=None
    ):
        self._r = r
        self._T = T
        self._dt = dt
        self._prices = prices
        self._asset_names = asset_names
        self._s0 = s0
        self._sigma = sigma
        self._cor = cor

        if prices is None:
            items = [s0, sigma, cor]
            if any([i is None for i in items]):
                raise ValueError(
                    "Must pass an array of prices to the constructor or provide optional \
                    arguments s0, sigma and cor."
                )

    def fit(self):
        """
        Method to fit the simulation parameters to the class. Not really needed for
        a GBM simulation, but included for consistency with other simulation classes.
        """
        # function not needed for this class since
        # all parameters are passed in the constructor
        # and there is nothing to fit.

        if self._prices is not None:
            prices = _pd.DataFrame(self._prices)
            returns = (_np.log(prices) - _np.log(prices.shift())).dropna()
            self._s0 = prices.iloc[-1, :]
            self._sigma = returns.std() * _np.sqrt(1 / self._dt)
            self._cor = returns.corr()

    def simulate(self, sims=1000):

        self._sims = simGBM_MV(
            self._s0, self._r, self._sigma, self._T, self._dt, cor=self._cor, sims=sims
        )


class MVOU(_MVSIM):
    """
    Class for simulating a multivariate OU process for a portfolio of assets

    Parameters
    ----------
    s0 : array-like[float], optional
        Initial price of each asset in the portfolio.
        Required if prices not passed.
    mu : float, int or pandas Series, optional
        Mean that the function will revert to. Can be either a scalar value (i.e. 5) or a pandas series for a
        time dependent mean. If array-like, it must be the same length as T/dt (i.e. the number of periods).
        Required if prices not passed.
    theta : float, optional
        Mean reversion rate, higher number means it will revert slower.
        Required if prices not passed.
    sigma : array-like[float], optional
        Volatility of each asset in the portfolio.
        Required if prices not passed.
    T : float
        Time to maturity of the portfolio.
    dt : float
        Time step for the simulation.
    cor : array-like[float]
        Correlation matrix for the assets in the portfolio.
    prices : array-like[float], optional
        Matrix or dataframe of asset prices of size (p x N) where p is the number
        of time steps and N is the number of assets. Used to calibrate model using fit
        method if s0, mu, theta, sigma and cor not passed.
    asset_names : list[str]
        List of strings to use as asset names in the output dataframe.

    Example
    -------
    >>> import risktools as rt
    >>> ou = rt.MVOU(
            s0=[5,5,5],
            mu=[4,4,4],
            theta=[2,2,2],
            sigma=[0.1,0.1,0.1],
            T=1,
            dt=1/252,
            cor=[[1,0.5,0.5],[0.5,1,0.5],[0.5,0.5,1]],
            asset_names=['A','B','C']
        )
    >>> ou.fit()
    >>> ou.simulate()
    >>> ou.plot_efficient_frontier()
    """

    def __init__(
        self,
        T,
        dt,
        s0=None,
        mu=None,
        theta=None,
        sigma=None,
        cor=None,
        prices=None,
        asset_names=None,
    ):
        self._mu = mu
        self._theta = theta
        self._T = T
        self._dt = dt
        self._prices = prices
        self._asset_names = asset_names
        self._s0 = s0
        self._sigma = sigma
        self._cor = cor

        if prices is None:
            items = [s0, mu, theta, sigma, cor]
            if any([i is None for i in items]):
                raise ValueError(
                    "Must pass an array of prices to the constructor or provide optional \
                    arguments s0, mu, theta, sigma and cor."
                )

    def fit(self, log_price=False, method="OLS", verbose=False, s0=None):
        """
        Method to fit the simulation parameters to the class.

        Parameters
        ----------
        log_price : bool
            If True, the spread is assumed to be log prices and the log of the spread is taken.
            Default is False.
        method : ['OLS', 'MLE']
            Method to use for parameter estimation. Default is 'OLS'.
        verbose : bool
            If True, prints the estimated parameters. Only used is method is OLS.
            Default is False.
        s0 : array-like[float]
            Of size N (number of assets) to override the last value for asset prices
            of the passed prices dataframe in the object initializations. Mostly for
            testing and teaching purposes. By default None.
        """
        # Only relevent if prices not passed to constructor
        if self._prices is not None:
            prices = _pd.DataFrame(self._prices)
            returns = (_np.log(prices) - _np.log(prices.shift())).dropna()

            self._s0 = prices.iloc[-1, :] if s0 is None else s0

            self._params = fitOU_MV(
                prices, self._dt, log_price=log_price, method=method, verbose=verbose
            )
            self._cor = returns.corr()
        else:
            self._params = _pd.DataFrame()
            self._params["mu"] = self._mu
            self._params["annualized_sigma"] = self._sigma
            self._params["theta"] = self._theta
            self._params = self._params.T

        if self._asset_names is not None:
            self._params.columns = self._asset_names

    def simulate(self, sims=1000):

        self._sims = self._sims = simOU_MV(
            self._s0,
            self._params.loc["mu", :],
            self._params.loc["theta", :],
            self._T,
            self._dt,
            self._params.loc["annualized_sigma", :],
            self._cor,
            sims=sims,
        )



import pandas as _pd
from scipy.optimize import linprog as _linprog


def refineryLP(crude_inputs, product_outputs, return_all=False):
    """
    Refinery optimization LP model
    
    Parameters
    ----------
    crude_inputs : DataFrame
    
    product_outputs : DataFrame

    Returns
    -------

    Examples
    --------
    >>> import risktools as rt
    >>> crudes = rt.data.open_data('ref_opt_inputs')
    >>> products = rt.data.open_data('ref_opt_outputs')
    >>> rt.refineryLP(crudes, products)
    """

    crudes = crude_inputs.copy()
    products = product_outputs.copy()
    crudes = crudes.set_index("info")

    gpw = _pd.DataFrame(
        dict(
            element=["gross_product_worth", "crude_cost", "processing"],
            light_sweet=[
                (products.prices * products.LightSweet_yield).sum(),
                crudes.loc["price", "LightSweet"],
                crudes.loc["processing_fee", "LightSweet"],
            ],
            heavy_sour=[
                (products.prices * products.HeavySour_yield).sum(),
                crudes.loc["price", "HeavySour"],
                crudes.loc["processing_fee", "HeavySour"],
            ],
        )
    )

    gpw = gpw[["light_sweet", "heavy_sour"]].sum()
    constraints = products[
        ["product", "LightSweet_yield", "HeavySour_yield", "max_prod"]
    ]

    out = _linprog(-gpw, A_ub=constraints.iloc[:, [1, 2]], b_ub=constraints.iloc[:, 3])

    if return_all == False:
        return dict(profit=-out["fun"], slate=out["x"])
    else:
        return out




import pandas as _pd
import requests as _requests
import urllib as _urllib
import re as _re
import io as _io


def get_prices(
    username,
    password,
    feed="CME_NymexFutures_EOD",
    codes=["CL9Z", "CL0F", "CL0M"],
    start_dt="2019-01-01",
    end_dt=None,
    intraday=False,
):
    """
    function to get prices from the Morningstar data API. The following feeds are supported

        * CME_CbotFuturesEOD and CME_CbotFuturesEOD_continuous
        * CME_NymexFutures_EOD and CME_NymexFutures_EOD_continuous
        * CME_NymexOptions_EOD
        * CME_CmeFutures_EOD and CME_CmeFutures_EOD_continuous
        * CME_Comex_FuturesSettlement_EOD and CME_Comex_FuturesSettlement_EOD_continuous
        * LME_AskBidPrices_Delayed
        * SHFE_FuturesSettlement_RT
        * ICE_EuroFutures and ICE_EuroFutures_continuous
        * ICE_NybotCoffeeSugarCocoaFutures and ICE_NybotCoffeeSugarCocoaFutures_continuous
        * CME_STLCPC_Futures
        * CFTC_CommitmentsOfTradersCombined. Requires multiple keys. Separate them by a space e.g. "N10 06765A NYME 01".
        * Morningstar_FX_Forwards. Requires multiple keys. Separate them by a space e.g. "USDCAD 2M".
        * ERCOT_LmpsByResourceNodeAndElectricalBus
        * PJM_Rt_Hourly_Lmp
        * AESO_ForecastAndActualPoolPrice

    Parameters
    ----------
    username : str
        Morningstar API username
    password : str
        Morningstar API password
    feed : str
        API feed to get data from, by default "CME_NymexFutures_EOD"
    codes : list[tuple[str]]
        either a string ticker code or a list of ticker codes or a futures contract type to return, by default ["CL9Z", "CL0F", "CL0M"]. 
    start_dt : str | datetime, optional
        earliest date to return data from, by default "2019-01-01"
    end_dt : str | datetime, optional
        lastest date to return data from, by default None. If None, the function return everything from start_dt forward
    intraday : bool
        not implemented yet

    Examples
    --------
    >>> import risktools as rt
    >>> rt.get_prices(username=up['m*']['user'], password=up['m*']['pass'])

    Returns
    -------
    Pandas dataframe with prices
    """
    if isinstance(codes, list) == False:
        codes = [codes]
    s = _requests.Session()
    s.auth = (username, password)

    start_dt = _pd.to_datetime(start_dt)

    url = "https://mp.morningstarcommodity.com/lds/feeds/{}/ts?{}"

    df = _pd.DataFrame()

    for code in codes:
        p_dict = dict()
        p_dict["fromDateTime"] = start_dt.strftime("%Y-%m-%d")

        if end_dt is not None:
            end_dt = _pd.to_datetime(end_dt)
            p_dict["toDateTime"] = end_dt.strftime("%Y-%m-%d")

        if feed in [
            "CME_NymexFutures_EOD",
            "CME_NymexOptions_EOD",
            "CME_CbotFuturesEOD",
            "CME_CmeFutures_EOD",
            "ICE_EuroFutures",
            "ICE_NybotCoffeeSugarCocoaFutures",
            "CME_Comex_FuturesSettlement_EOD",
            "LME_AskBidPrices_Delayed",
            "SHFE_FuturesSettlement_RT",
        ]:
            p_dict["Symbol"] = code
        elif feed in [
            "CME_NymexFutures_EOD_continuous",
            "CME_CmeFutures_EOD_continuous",
            "ICE_EuroFutures_continuous",
            "ICE_NybotCoffeeSugarCocoaFutures_continuous",
            "CME_Comex_FuturesSettlement_EOD_continuous",
            "CME_CbotFuturesEOD_continuous",
        ]:
            p_dict["Contract"] = code
        elif feed in ["CME_STLCPC_Futures"]:
            p_dict["product"] = code
        elif feed in ["CFTC_CommitmentsOfTradersCombined"]:
            fcode = _re.sub("[^\w]", " ", code).split()  # get rid of any punctuation
            p_dict["cftc_subgroup_code"] = fcode[0]
            p_dict["cftc_contract_market_code"] = fcode[1]
            p_dict["cftc_market_code"] = fcode[2]
            p_dict["cftc_region_code"] = fcode[3]
            p_dict["cols"] = fcode[4]
        elif feed in ["Morningstar_FX_Forwards"]:
            fcode = _re.sub("[^\w]", " ", code).split()  # get rid of any punctuation
            p_dict["cross_currencies"] = fcode[0]
            p_dict["period"] = fcode[1]
        elif feed in ["ERCOT_LmpsByResourceNodeAndElectricalBus"]:
            p_dict["SettlementPoint"] = code
        elif feed in ["PJM_Rt_Hourly_Lmp"]:
            p_dict["pnodeid"] = code
        elif feed in ["AESO_ForecastAndActualPoolPrice"]:
            p_dict["market"] = code
        elif feed in ["LME_MonthlyDelayed_Derived"]:
            code_s = code.split()
            p_dict["Root"] = code_s[0]
            p_dict["DeliveryStart"] = code_s[1]
            p_dict["DeliveryEnd"] = code_s[2]
        else:
            raise ValueError("feed not recognized")

        params = _urllib.parse.urlencode(p_dict)
        r = s.get(url.format(feed, params),)

        tf = _pd.read_csv(_io.StringIO(r.content.decode("utf-8"))).set_index("Date")
        tf.index = _pd.to_datetime(tf.index)
        tf.columns = tf.columns.str.replace(
            "\(([^)]*)\)", "", regex=True
        )  # clean up columns by removing ticker/code names
        tf = _pd.concat([tf], keys=[code])

        # df = df.append(tf)
        df = _pd.concat([df, tf], axis=0)

    return df


def get_curves(
    username,
    password,
    feed="Crb_Futures_Price_Volume_And_Open_Interest",
    contract_roots=["CL", "BG"],
    fields=["Open", "High", "Low", "Close"],
    date=None,
):
    """
    function to get forward curves from the Morningstar data API. The following feeds are supported

    * Crb_Futures_Price_Volume_And_Open_Interest
    * CME_NymexFuturesIntraday_EOD
    * ICE_EuroFutures
    * ICE_EuroFutures_continuous

    Parameters
    ----------
    username : str
        Morningstar API username
    password : str
        Morningstar API password
    feed : str
        API feed to get data from, by default "CME_NymexFutures_EOD"
    contract_roots : list[tuple[str]]
        either a string contract root code (i.e. "CL" for NYMEX WTI) or a list roots to return, by default ["CL", "BG"]. 
    date : str | datetime, optional
        Date of curve to pull. By default None which causes function to return data for today.

    Examples
    --------
    >>> import risktools as rt
    >>> rt.get_curves(
        username=username, 
        password=password, 
        feed='Crb_Futures_Price_Volume_And_Open_Interest', 
        contract_roots=['CL', 'BG'], 
        fields=['Open', 'High', 'Low', 'Close'], 
        date=None
        )
    """
    if isinstance(contract_roots, list) == False:
        contract_roots = [contract_roots]
    s = _requests.Session()
    s.auth = (username, password)

    if date is not None:
        date = _pd.to_datetime(date)
    else:
        date = _pd.Timestamp.now().date()

    url = "https://mp.morningstarcommodity.com/lds/feeds/{}/curve?{}"

    df = _pd.DataFrame()

    for root in contract_roots:
        p_dict = dict()

        if feed in [
            "Crb_Futures_Price_Volume_And_Open_Interest",
            "CME_NymexFuturesIntraday_EOD",
            "ICE_EuroFutures",
            "ICE_EuroFutures_continuous",
        ]:
            p_dict["root"] = root
        else:
            raise ValueError("feed not recognized")

        p_dict["cols"] = ",".join(fields)
        p_dict["date"] = date.strftime("%Y-%m-%d")
        # p_dict["out"] = "json"

        # get data from API
        params = _urllib.parse.urlencode(p_dict)
        r = s.get(url.format(feed, params),)

        # format df
        tf = _pd.DataFrame(r.json())
        tf["code"] = tf["keys"].apply(lambda x: x[0]["vendorcode"])
        tf = tf.drop("keys", axis=1)

        tf.value = _pd.to_numeric(tf.value)
        tf.expirationDate = _pd.to_datetime(tf.expirationDate)

        tf["contract"] = root
        tf = tf[["contract", "code", "expirationDate", "col", "value"]]
        tf = tf.rename({"col": "type"}, axis=1)

        tf = (
            tf.set_index(["contract", "code", "expirationDate", "type"])
            .unstack()
            .droplevel(0, 1)
            .reset_index()
            .sort_values("expirationDate")
        )

        tf["rownum"] = tf.index + 1
        tf["contract"] += tf.rownum.astype(str).str.rjust(2, "0")
        tf = tf.drop("rownum", axis=1)

        tf = tf[["contract", "code", "expirationDate", *fields]]
        tf.columns.name = ""

        if len(contract_roots) > 1:
            tf["root"] = root
            tf = tf[
                [
                    "root",
                    "contract",
                    "code",
                    "expirationDate",
                    "Open",
                    "High",
                    "Low",
                    "Close",
                ]
            ]

        # df = df.append(tf)
        df = _pd.concat([df,tf], axis=0)

    return df


if __name__ == "__main__":
    import json

    with open("../../user.json") as jfile:
        userfile = jfile.read()

    up = json.loads(userfile)

    username = up["m*"]["user"]
    password = up["m*"]["pass"]
    # df = get_prices(
    #     username=username,
    #     password=password,
    #     feed="CME_NymexFutures_EOD",
    #     codes=["@CL0Z", "@CL21Z"],
    #     start_dt="2019-08-26",
    # )

    # df = get_prices(feed="CME_NymexFutures_EOD",codes="@CL0Z",start_dt="2019-08-26",username = username, password = password)
    # df = get_prices(feed="CME_NymexFutures_EOD",codes="@CL21Z",start_dt="2019-08-26",username = username, password = password)
    # df = get_prices(
    #     feed="CME_NymexFutures_EOD_continuous",
    #     codes=["CL_006_Month", "CL_007_Month"],
    #     start_dt="2019-08-26",
    #     username=username,
    #     password=password,
    # )
    # df = get_prices(feed="CME_NymexOptions_EOD",codes="@LO21ZP4000",start_dt="2020-03-15",username = username, password = password)
    # df = get_prices(feed="CME_CbotFuturesEOD",codes="C0Z",start_dt="2019-08-26",username = username, password = password)
    # df = get_prices(feed="CME_CbotFuturesEOD_continuous",codes="ZB_001_Month",start_dt="2019-08-26",username = username, password = password)
    # df = get_prices(feed="CME_CmeFutures_EOD_continuous",codes="HE_006_Month",start_dt="2019-08-26",username = username, password = password)
    # df = get_prices(feed="Morningstar_FX_Forwards",codes="USDCAD 2M",start_dt="2019-08-26",username = username, password = password)
    # df = get_prices(feed="CME_CmeFutures_EOD",codes="LH0N",start_dt="2019-08-26",username = username, password = password)
    # df = get_prices(feed="CME_CmeFutures_EOD_continuous",codes="HE_006_Month",start_dt="2019-08-26",username = username, password = password)
    # df = get_prices(feed="ICE_EuroFutures",codes="BRN0Z",start_dt="2019-08-26",username = username, password = password)
    # df = get_prices(feed="ICE_EuroFutures_continuous",codes="BRN_001_Month",start_dt="2019-08-26",username = username, password = password)
    # df = get_prices(feed="ICE_NybotCoffeeSugarCocoaFutures",codes="SB21H",start_dt="2019-08-26",username = username, password = password)
    df = get_prices(
        feed="ICE_NybotCoffeeSugarCocoaFutures_continuous",
        codes="SF_001_Month",
        start_dt="2019-08-26",
        username=username,
        password=password,
    )

    print(df)



from .__init__ import *
from . import data
import pandas as pd
import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots as _make_subplots
from plotly.colors import DEFAULT_PLOTLY_COLORS
from math import ceil as _ceil
from ._morningstar import *
from statsmodels.tsa.seasonal import STL as _STL
from ._pa import *

from ._main_functions import get_eia_df, infer_freq
from ._cullenfrey import describe_distribution as _desc_dist
import matplotlib.pyplot as _plt
import seaborn as _sns
import arch as _arch
from matplotlib.pyplot import cm as _cm


def chart_zscore(df, freq=None, output="zscore", chart="seasons", **kwargs):
    """
    Function for calculating and plotting the seasonal decomposition and zscore of seasonal data
    Z-score is computed on the residuals conditional on their seasonal period. Beware that most
    seasonal charts in industry (i.e. NG Storage) is not detrended so results once you apply
    STL decomposition will vary from the unadjusted seasonal plot

    Parameters
    ----------
    df : Series
        Series with date and a values column. Datetime index must have a frequency defined (such as 'D', 'B', 'W', 'M' etc...)
    freq : str, optional
        Resampling frequency, by default None. Resamples using mean if downsampling. Not designed for
        upsampling.
    output : ['stl','zscore','seasonal']
        'stl' - for seasonal decomposition object. Run the object method .plot() to plot the STL decomposition.
        Also use the object attributes "observed", "resid", "seasonal", "trend" and "weights" to see components
        and calculate stats.
        'zscore' - return residuals of szore
        'seasonal' - for standard seasonal chart
        By default 'zscore'
    chart : str, optional
        by default 'seasons'

    Returns
    -------
    'stl' - Statsmodels STL object
    'seasonal' - Plotly figure object
    'zscore' - Plotly figure object

    Examples
    --------
    >>> df = rt.data.open_data('eiaStocks')
    >>> df = df.loc[df.series=='NGLower48',['date','value']].set_index('date')['value']
    >>> df = df.resample('W-FRI').mean()
    >>> stl = rt.chart_zscore(df, freq='M')
    >>> stl.show()
    >>> stl = rt.chart_zscore(df, output='seasonal')
    >>> stl.show()
    >>> stl = rt.chart_zscore(df, output='zscore')
    >>> stl.show()
    """

    # either resample or figure out frequency of the series
    if freq is not None:
        df = df.resample(freq).mean()
    elif df.index.freq is None:
        df.index.freq = pd.infer_freq(df.index[0:10])

    if str(df.index.freqstr)[0] == "M":
        seasonal = 13
    elif str(df.index.freqstr)[0] == "W":
        seasonal = 53
    else:
        seasonal = 7

    stl = _STL(df, seasonal=seasonal, seasonal_deg=0, robust=False).fit()

    if output == "stl":
        return stl
    elif output == "seasonal":
        df = pd.concat([df], keys=["value"], axis=1)
        df["year"] = df.index.year
        df.index = df.index + pd.DateOffset(year=df.year.max())
        df = df.set_index("year", append=True).sort_index()
        df = df.unstack(level=-1).droplevel(0, 1)

        fig = go.Figure()
        for c in df.columns:
            if c == df.index.year.max():
                mode = "lines+markers"
            else:
                mode = "lines"
            fig.add_trace(
                go.Scattergl(
                    x=df[c].dropna().index, y=df[c].dropna(), name=c, mode=mode
                )
            )
        fig.update_layout(xaxis_title="freq", yaxis_title="value")
        fig.update_layout(
            xaxis={
                "tickformat": "%b",
                # tickmode': 'auto',
                "nticks": 13,  # [where value is the max # of ticks]
                # 'tick0': '01', # [where value is the first tick]
                # 'dtick': 1, # [where value is the step between ticks]
            }
        )
        if kwargs is not None:
            fig.update_layout(**kwargs)

        return fig
    elif output == "zscore":
        zs = pd.DataFrame()

        # first calculate mean and std of the residuals by frequency (i.e. by week, month etc...) for all years
        zs["resid"] = stl.resid
        zs["year"] = zs.index.year
        tt = stl.resid.index.to_series()
        zs["per"] = (
            tt.groupby(tt.dt.year).cumcount() + 1
        )  # elegant way to get weeknum, monthnum, quarter etc... regardless of freq
        ag = zs.groupby("per").resid.agg(["mean", "std"]).reset_index()

        zs = (
            zs.reset_index().merge(ag, left_on="per", right_on="per").set_index("date")
        )  # use reset index or it drops the datetime index of zs
        zs["z_score"] = (zs.resid - zs["mean"]) / zs["std"]

        fig = go.Figure(
            data=[
                go.Bar(
                    x=zs.index,
                    y=zs.z_score,
                    marker_color=zs.z_score,
                    marker_colorscale=["red", "orange", "green", "orange", "red"],
                    marker_colorbar=dict(title="Z-Score"),
                )
            ]
        )

        fig.update_layout(
            title="Z-Score of Seasonally Adjusted Residuals",
            xaxis_title="date",
            yaxis_title="Z Score",
        )
        if kwargs is not None:
            fig.update_layout(**kwargs)

        return fig


def chart_eia_sd(market, key, start_dt="2010-01-01", output="chart", **kwargs):
    """
    Function for plotting and returning data from the EIA Supply & Demand Balances for
    refined oil products such as mogas, diesel, jet and resid.

    Parameters
    ----------
    market : ['mogas', 'diesel', 'jet', 'resid']
        Refined product type to build the S&D balance for
    key : str
        EIA API key
    start_dt : str | datetime
        Starting date for S&D balance data
    output : ['chart','data']
        Output as a chart or return data as a dataframe, by default 'chart'

    Returns
    -------
    pandas dataframe or a plotly figure object

    Examples
    --------
    >>> import risktools as rt
    >>> fig = rt.chart_eia_sd('mogas', up['eia'])
    >>> fig.show()
    """
    eia = data.open_data("tickers_eia")
    eia = eia[eia.sd_category == market]

    df = get_eia_df(eia.tick_eia.to_list(), key=key)
    df = df.merge(
        eia[["tick_eia", "category"]], left_on=["series_id"], right_on=["tick_eia"]
    ).drop("tick_eia", axis=1)

    df.date = pd.to_datetime(df.date)
    df = df.set_index("date").sort_index()

    # create list of plotly figure objects using repeating calls to the
    # five_year_plot function to loop through later to create subplot
    figs = []
    for c in df.category.unique():
        tf = df.loc[df.category == c, "value"]
        figs.append(chart_five_year_plot(tf, title=c))

    # calc shape of final subplot
    n = len(df.category.unique())
    m = 2
    n = _ceil(n / m)
    fig = _make_subplots(n, m, subplot_titles=(" ", " ", " ", " ", " ", " "),)

    # Copy returns figures from five_year_plot to a single subplot figure
    a = 1
    b = 1
    for i, _ in enumerate(figs):
        for j, _ in enumerate(figs[i]["data"]):
            fig.add_trace(figs[i]["data"][j], row=a, col=b)
            fig["layout"]["annotations"][(a - 1) * 2 + b - 1]["text"] = figs[i][
                "layout"
            ]["title"]["text"]

        # copy xaxis nticks and tickformat to subplots so that it keeps that formatting.
        # if they don't exist, pass
        try:
            fig["layout"][f"xaxis{i+1}"]["nticks"] = figs[i]["layout"]["xaxis"][
                "nticks"
            ]
            fig["layout"][f"xaxis{i+1}"]["tickformat"] = figs[i]["layout"]["xaxis"][
                "tickformat"
            ]
        except:
            pass

        if b == m:
            b = 1
            a += 1
        else:
            b += 1

    fig.update_layout(showlegend=False)
    # return figs
    if output == "chart":
        return fig
    else:
        return df


def chart_five_year_plot(df, resample_freq=None, **kwargs):
    """
    Function to output a 5 yr range plot for commodity prices

    Parameters
    ----------
    df : Series
        Series or single column dataframe with a datetime index
    resample_freq : str
        Used to resample data. See Pandas documentation for valid freq types.
        Uses the resampled mean. By default None.

    Returns
    -------
    Plotly figure object

    Example
    -------
    >>> import risktools as rt
    >>> df = rt.data.open_data('dfwide')
    >>> rt.chart_five_year_plot(df['CL01'])
    """
    df = df.copy()
    if isinstance(df, pd.Series):
        df = pd.DataFrame({"value": df})

    if resample_freq is None:
        freq = pd.infer_freq(df.index[-10:])
    else:
        freq = resample_freq

    df["year"] = df.index.year
    start_dt = str(df.index.year.max() - 5) + "-01-01"

    cy = df.index.year.max()
    py = cy - 1

    df = df.sort_index().loc[start_dt:, :]
    fy = df.index.year.min()

    df.index = df.index + pd.DateOffset(year=cy)
    df = df.set_index("year", append=True).sort_index().unstack().droplevel(0, 1)
    df = df.resample(freq).mean()

    # shift back columns where the first row is NaN for offset weeks
    for c in df.columns:
        if np.isnan(df[c].iloc[0]):
            df[c] = df[c].shift(-1)
    df = df.dropna(thresh=2, axis=0)
    mf = df[range(fy, cy)].resample(freq).mean()

    fig = go.Figure()
    fig.add_trace(
        go.Scatter(
            x=mf.min(axis=1).index, y=mf.min(axis=1), line_color="lightgrey", name="min"
        )
    )
    fig.add_trace(
        go.Scatter(
            x=mf.max(axis=1).index,
            y=mf.max(axis=1),
            line_color="lightgrey",
            name="max",
            fill="tonexty",
        )
    )

    fig.add_trace(
        go.Scatter(
            x=df[cy].dropna().index,
            y=df[cy].dropna(),
            name=str(cy),
            line_color="#EF553B",
        )
    )
    fig.add_trace(
        go.Scatter(
            x=df[py].dropna().index,
            y=df[py].dropna(),
            name=str(py),
            line_color="royalblue",
        )
    )
    fig.add_trace(
        go.Scatter(
            x=mf.mean(axis=1).index,
            y=mf.mean(axis=1),
            name="5 yr mean",
            line_dash="dash",
            line_color="black",
        )
    )

    fig.update_layout(xaxis={"tickformat": "%b", "nticks": 13})
    fig.update_layout(**kwargs)

    return fig


def chart_eia_steo(key, start_dt=None, market="globalOil", output="chart", **kwargs):
    """
    Function to plot the global oil supply and demand balance using the US EIA data

    Parameters
    ----------
    key : str
        Your EIA API key
    start_dt : str | datetime
        A string in the format of 'YYYY-mm-dd' or a datetime object
    market : ['globalOil']
        What market to build the S&D balance for. For now, only globalOil working.
    output : ['chart', 'data'], optional
        What kind of output to return, by default 'chart'
    **kwargs
        keyword arguments to pass to plotly.graph_objects.figure.update_layout function to modify figure

    Returns
    -------
    output = 'chart'
        Plotly figure object
    output = 'data'
        dataframe

    Examples
    --------
    >>> import risktool as rt
    >>> fig = rt.chart_eia_steo(up['eia'])
    >>> fig.show()
    """
    if market == "globalOil":
        tickers = {
            "STEO.PAPR_NONOPEC.M": "SupplyNOPEC",
            "STEO.PAPR_OPEC.M": "SupplyOPEC",
            "STEO.PATC_WORLD.M": "Demand",
            "STEO.T3_STCHANGE_WORLD.M": "Inv_Change",
        }

    df = get_eia_df(list(tickers.keys()), key=key)
    df["name"] = df["series_id"].map(tickers)
    df = (
        df[["date", "value", "name"]]
        .set_index(["date", "name"])
        .sort_index()
        .unstack()
        .droplevel(0, 1)
        .dropna()
    )
    df["Supply"] = df.SupplyNOPEC + df.SupplyOPEC
    df.Inv_Change *= -1
    df = df.drop(["SupplyNOPEC", "SupplyOPEC"], axis=1)

    if start_dt is not None:
        df = df[start_dt:]

    if output == "data":
        return df
    else:
        fig = go.Figure()
        for c in df.columns:
            fig.add_trace(go.Scatter(x=df.index, y=df[c], mode="lines", name=c))
        fig.update_layout(
            title=f"US EIA {market} Supply and Demand Balance",
            xaxis_title="date",
            yaxis_title="millions of barrels per day",
        )
        if kwargs is not None:
            fig.update_layout(**kwargs)
        return fig


def chart_perf_summary(df, geometric=True, title=None):
    """
    Function to plot the cumulative performance and drawdowns of
    multiple assets

    Parameters
    ----------
    df : DataFrame | Series
        wide dataframe with a datetime index and assets as columns or a
        series with a datetime index for a univariate chart
    geometric : bool
        True to plot geometric returns, False to plot absolute
    title : str
        Title for plot


    >>> import risktools as rt
    >>> df = rt.data.open_data('dfwide')
    >>> df = df[['CL01', 'CL12', 'CL36']]
    >>> df = rt.returns(df, period_return=1)
    >>> rt.chart_perf_summary(df, geometric=False, title="Cummulative Returns and Drawdowns")
    """
    df = df.copy()
    if geometric == True:
        ret = df.add(1).cumprod()
    else:
        ret = df.cumsum()

    dd = drawdowns(df, geometric=geometric)

    cols = DEFAULT_PLOTLY_COLORS

    fig = _make_subplots(
        rows=2, cols=1, subplot_titles=("Cummulative Returns", "Drawdowns")
    )
    for i, c in enumerate(ret.columns):
        fig.add_trace(
            go.Scatter(
                x=ret.index,
                y=ret[c],
                name=c,
                line=dict(width=2, color=cols[i]),
                legendgroup="a",
            ),
            row=1,
            col=1,
        )

    for i, c in enumerate(dd.columns):
        fig.add_trace(
            go.Scatter(
                x=dd.index,
                y=dd[c],
                name=c,
                line=dict(width=2, color=cols[i]),
                legendgroup="b",
                showlegend=False,
            ),
            row=2,
            col=1,
        )

    fig.update_layout(title_text=title)
    return fig


def chart_forward_curves(
    df,
    code=None,
    cmdty=None,
    resample_freq=None,
    skip=5,
    curve_len=365 * 2,
    yaxis_title="$/BBL",
    **kwargs,
):
    """
    Chart forward curves of a futures contract strip

    Parameters
    ----------
    df : DataFrame
        pandas DataFrame with a datetime index and columns representing the futures contract, ordered by most recent expiry
    code : str, optional
        commodity code to use to filter
    cmdty : str, optional
        expiry_table code to use to get expiry dates to plot forward curves. See data.open_data('expiry_data') for allowable
        values, using the column 'cmdty'
    resample_freq : str, optional
        pandas resample frequency type to downsample by. Assumes that new freq of lower periodicity than old freq. Uses mean to downsample
    skip : int
        number of days between forward curves, must be 1 or greater
    curve_len : int
        how long the draw the curve on the plot, in days. Required if cmdty is not given.
    uom : str
        Unit of Measure to show on the final chart y-axis
    ***kwargs
        keyword arguments to pass to fig.update_layout function for Plotly figures. Mainly
        to adjust figure layout.

    Returns
    -------
    Chart of forward curves varying with time

    Examples
    --------
    >>> import risktools as rt
    >>> df = rt.data.open_data('dfwide')
    >>> rt.chart_forward_curves(df, 'HO', yaxis_title='$/g', skip=10)
    >>> rt.chart_forward_curves(df, 'HO', cmdty='cmeulsd', yaxis_title='$/g', skip=2)
    """
    df = df.copy()
    fig = go.Figure()

    # if cmdty is defined, then use the expiry_table table to extract the expiry
    # schedule to pass to the _plot_crv function. Otherwise pass None to the function
    # so that it uses the defined time step
    if isinstance(cmdty, str):
        exp = data.open_data("expiry_table")
        exp = exp.loc[exp.cmdty == cmdty, "Last_Trade"]
    else:
        exp = None

    if code is not None:
        df = df[df.columns[df.columns.str.startswith(code)]]

    if resample_freq is not None:
        df = df.resample(resample_freq).mean()

    def _plot_crv(df, date, d=curve_len, exp=exp):
        """
        Convert series index to align with front contract dates with the curve spanning 400 days thereafter
        df : array-like
            dataframe with all contracts to be plotted
        d : int
            number of days that the curve should span

        Returns
        -------
        pandas series to plot with modified datetime index
        """
        df = df.copy()
        df.columns = df.columns.str.replace("[^0-9]", "", regex=True).astype(int)
        se = df.loc[date, :]

        n = len(se)
        days_per_step = _ceil(d / n)

        if exp is None:
            idx = []
            for i, r in se.items():
                idx.append(
                    pd.to_datetime(date) + pd.DateOffset((i - 1) * days_per_step)
                )
        else:
            idx = exp[exp > date]
            idx = idx.iloc[: len(se)]

        se.index = pd.Index(idx)

        return se

    cnt = 0
    for i, r in df.iterrows():
        if cnt % skip == 0:
            crv = _plot_crv(df, i)
            fig.add_trace(
                go.Scattergl(
                    x=crv.index,
                    y=crv,
                    line=dict(width=1, dash="dash"),
                    legendgroup="b",
                    showlegend=False,
                    mode="lines",
                )
            )
        cnt += 1

    # add front month trace
    fig.add_trace(
        go.Scattergl(
            x=df.index,
            y=df.iloc[:, 0],
            name="Front Contract",
            legendgroup="a",
            line_color="royalblue",
        )
    )

    # leave this here before **kwargs so that users can adjust
    fig.update_layout(width=800, height=800)

    fig.update_layout(
        dict(title=f"{code} Forward Curves", yaxis_title=yaxis_title, **kwargs,)
    )

    # Add range slider
    fig.update_layout(
        xaxis=dict(
            rangeslider=dict(visible=True, autorange=True, thickness=0.05), type="date",
        )
    )

    return fig


def chart_pairs(df, title="Time Series Pairs Plot", **kwargs):
    """
    Pairwise scatter matrix plot for timeseries

    Parameters
    ----------
    df : DataFrame
        pandas DataFrame with a datetime index and columns representing the futures contract, ordered by most recent expiry
    title : str, optional
        Chart title, by default "Time Series Pairs Plot"
    **kwargs
        keyword arguments to pass to plotly.graph_objects.Figure.update_layout function

    Examples
    --------
    >>> import risktools as rt
    >>> df = rt.data.open_data('dfwide')
    >>> df = df[['CL01', 'NG01', 'HO01', 'RB01']].dropna()
    >>> rt.chart_pairs(df)
    """
    dt_idx = df.index.name
    df = df.reset_index().copy()

    dims = []
    for c, i in df.items():
        dims.append(dict(label=c, values=df[c],))

    fig = go.Figure()
    fig.add_trace(
        go.Splom(
            dimensions=dims,
            showupperhalf=False,
            marker=dict(color=df[dt_idx].astype(int), colorscale="Portland"),
            diagonal_visible=False,
        )
    )
    fig.update_layout(width=1000, height=1000, title=title)
    if kwargs is not None:
        fig.update_layout(**kwargs)

    return fig


def chart_spreads(
    username,
    password,
    pairs,
    days_from_expiry=200,
    feed=None,
    output="chart",
    start_dt="2012-01-01",
    **kwargs,
):
    """
    Chart commodity price spreads given as a list of tuple with two commodity codes and a name

    Parameters
    ----------
    username : str
        username for Morningstar API
    password : str
        password for Morningstar API
    pairs : List of Tuples with String
        A list of tuples of the format [(code_1, code_2, legend_label), (..,..,..)]. The spread calculated is then
        (code_1 - code_2). For example to plot the March/April time spread for NYMEX ULSD for 2014 & 2019, it would
        be [('@HO4H', '@HO4J', '2014'), ('@HO9H', '@HO9J', '2019')]
    days_from_expiry : int
        Days from expiry of the earliest contract to start the plot from (i.e. 200 days). By default 200
    feed : str
        feed to use to get data from Morningstar
    start_dt : str
        Start date for the data retrival as a string, in the format "yyyy-mm-dd"
    output : ['data', 'chart']
        whether to return data or a chart, by default 'chart'
    **kwargs
        Arguments to pass to the plotly update_layout function to control the layout elements of the final figure. You
        can also modify the returned figure object if required.

    Returns
    -------
    a dataframe or a plotly figure object

    Examples
    --------
    >>> import risktools as rt
    >>> fig = rt.chart_spreads(
                username=up['m*']['user'],
                password=up['m*']['pass'],
                pairs=[('@HO4H', '@HO4J', '2014'), ('@HO9H', '@HO9J', '2019')],
                feed="CME_NymexFutures_EOD")
    >>> fig.show()
    """

    codes = []

    for c in pairs:
        codes.append(c[0])
        codes.append(c[1])

    df = get_prices(
        username=username, password=password, feed=feed, codes=codes, start_dt=start_dt
    )

    df = df.unstack(level=0).settlement_price

    res = pd.DataFrame()
    fig = go.Figure()

    for c in pairs:
        tmp = (df[c[0]] - df[c[1]]).dropna().reset_index(drop=True)[-days_from_expiry:]

        # reset index twice to bring in index number
        tmp = tmp.reset_index().reset_index()
        tmp.columns = ["days_from_exp", "date", "spread"]
        tmp.days_from_exp -= days_from_expiry
        tmp["year"] = c[2]
        # res = res.append(tmp)
        res = pd.concat([res, tmp], axis=0)

        if output == "chart":
            fig.add_trace(
                go.Scatter(
                    x=tmp.dropna().days_from_exp, y=tmp.dropna().spread, name=c[2]
                )
            )

    if kwargs is not None:
        fig.update_layout(**kwargs)

    if output == "chart":
        return fig
    else:
        return res


def dist_desc_plot(x, figsize=(10, 10)):
    """
    Provides a summary of returns' distribution

    Parameters
    ----------
    x : Pandas Series
        iterable of numeric returns

    Returns
    -------
    matplotlib Figure object

    Examples
    --------
    >>> import risktools as rt
    >>> df = rt.data.open_data('dflong')
    >>> x = df['BRN01'].pct_change().dropna()
    >>> rt.dist_desc_plot(x)     
    """
    _plt.figure(figsize=figsize)
    ax1 = _plt.subplot2grid((4, 2), (0, 0), rowspan=2)  # return dist
    ax2 = _plt.subplot2grid((4, 2), (2, 0), rowspan=2)  # cullenfrey
    ax3 = _plt.subplot(422)  # daily returns ts
    ax4 = _plt.subplot(424, sharex=ax3)  # garch vol
    ax5 = _plt.subplot(426, sharex=ax3)  # cum ret
    ax6 = _plt.subplot(428, sharex=ax3)  #

    # Returns Histogram
    _sns.histplot(x, ax=ax1, kde=True, stat="density")
    _sns.rugplot(x, ax=ax1)
    norm = np.random.normal(scale=x.std(), size=len(x) * 10)
    _sns.kdeplot(norm, ax=ax1, color="purple")
    ax1.set_xlabel("Returns")
    ax1.set_title("Return Distribution")

    # ax1.axvline(x.std() * -1.65, color="grey", ls="--", lw=1)
    ax1.axvline(np.percentile(x, 5), color="grey", ls="--", lw=1)
    _, ymax = ax1.get_ylim()
    # ax1.text(x.std() * -1.65 * 1.5, 0.8 * ymax, "95% VaR", rotation=90)
    ax1.text(np.percentile(x, 5) * 1.5, 0.8 * ymax, "95% VaR", rotation=90)

    # Cullen and Frey
    _desc_dist(x, boot=500, discrete=False, method="unbiased", ax=ax2)

    color = iter(_cm.rainbow(np.linspace(0, 1, 10)))
    c = next(color)

    # daily returns time series
    ax3.plot(x, c=next(color))
    ax3.set_title("Returns")

    # garch vol
    # *100 to scale it for model based on typical daily returns
    am = _arch.univariate.arch_model(x * 100, vol="Garch",)
    res = am.fit(disp="off").conditional_volatility / 100
    try:
        ann = np.sqrt(infer_freq(x, multiplier=True))
        text = "Annualized "
    except:
        ann = 1
        text = ""
    ax4.fill_between(res.index, res * ann, 0, color=next(color))
    ax4.set_title(f"Garch (1,0,1) {text}Volatility")

    # Cummulative Returns
    ax5.fill_between(x.index, x.cumsum(), 0, color=next(color))
    ax5.set_title("Cummulative Returns")

    # Drawdowns
    dd = drawdowns(x)
    ax6.fill_between(dd.index, dd, 0, color=next(color))
    ax6.set_title("Drawdowns")

    _plt.tight_layout()



import pandas as pd
import geopandas
import os
import json
import numpy as np

def norm_df(fn):
    rf = pd.DataFrame() # create results df
    
    # read json files with nested dataframes from R
    df = pd.read_json(fn)
    
    for c in df.columns:
        tmp = pd.json_normalize(df[c])

        tmp = pd.concat([tmp.set_index(tmp.columns[0])], keys=[c], axis=1).T
        rf = rf.append(tmp)

    rf.columns.name='specifications'
    rf.index = rf.index.set_names(['crude','cut'])
    return rf.sort_index()

def read_curves(fn):
    # open SwapCurve files (based on R S3 class Discount Curves) and convert to 
    # dictionary with nested arrays, dictionaries and one nested dataframe

    with open(fn) as f:
        dd = json.load(f)

    for key in dd.keys():
        if isinstance(dd[key],list) == True:
            dd[key] = np.array(dd[key])

    dd['table'] = pd.DataFrame(dd['table'])
    dd['table']['date'] = pd.to_datetime(dd['table']['date'], utc=True, unit='D')

    return dd

file_actions = {
    'cancrudeassays.json':{'date_fields':['YM'], 'load_func':pd.read_json},
    'cancrudeassayssum.json':{'date_fields':['YM'], 'load_func':pd.read_json},
    'cancrudeprices.json':{'date_fields':['YM'], 'load_func':pd.read_json},
    'crudeassaysBP.json':{'date_fields':None, 'load_func':norm_df},
    'crudeassaysXOM.json':{'date_fields':None, 'load_func':norm_df},
    'crudepipelines.geojson':{'date_fields':None, 'load_func':geopandas.read_file},
    'crudes.json':{'date_fields':None,'load_func':pd.read_json},
    'dflong.json':{'date_fields':['date'],'load_func':pd.read_json},
    'dfwide.json':{'date_fields':['date'],'load_func':pd.read_json},
    'df_fut.json':{'date_fields':['date'],'load_func':pd.read_json},
    'eiaStocks.json':{'date_fields':['date'],'load_func':pd.read_json},
    'eiaStorageCap.json':{'date_fields':['date'],'load_func':pd.read_json},
    'expiry_table.json':{'date_fields':['Last_Trade','First_Notice','First_Delivery','Last_Delivery'],'load_func':pd.read_json},
    'fizdiffs.json':{'date_fields':['date'],'load_func':pd.read_json},
    'holidaysOil.json':{'date_fields':['value'],'load_func':pd.read_json},
    'planets.json':{'date_fields':None,'load_func':pd.read_json},
    'productspipelines.geojson':{'date_fields':None, 'load_func':geopandas.read_file},
    'productsterminals.geojson':{'date_fields':None, 'load_func':geopandas.read_file},
    'ref.opt.inputs.json':{'date_fields':None,'load_func':pd.read_json},
    'ref.opt.outputs.json':{'date_fields':None,'load_func':pd.read_json},
    'refineries.geojson':{'date_fields':None, 'load_func':geopandas.read_file},
    'tickers_eia.json':{'date_fields':None,'load_func':pd.read_json},
    'tradeCycle.json':{'date_fields':['flowmonth','trade_cycle_end'],'load_func':pd.read_json},
    'tradeprocess.json':{'date_fields':['date'],'load_func':pd.read_json},
    'usSwapCurves.json':{'date_fields':None, 'load_func':read_curves},
    'usSwapCurvesPar.json':{'date_fields':None, 'load_func':read_curves},
    'usSwapIR.json':{'date_fields':['date'],'load_func':pd.read_json},
    'usSwapIRdef.json':{'date_fields':None,'load_func':pd.read_json},
}

path = os.path.dirname(__file__)
files = os.listdir(path)
files = [os.path.join(path, f) for f in files]

for fp in files:
    # assert fn in file_actions.keys(), f'{fn} not defined in file_actions dictionary'
    # get filename to pass to file_actions dict. Note that I use the full file path
    # to open the file
    fn = os.path.basename(fp)
    if fn in file_actions.keys():
        df = file_actions[fn]['load_func'](fp)

        if isinstance(df, pd.DataFrame):
            # convert "." to "_" in column names
            df.columns = df.columns.str.replace('.','_')

        # convert datetime fields
        if file_actions[fn]['date_fields'] is not None:
            for d in file_actions[fn]['date_fields']:
                df[d] = pd.to_datetime(df[d])

        var = fn.replace('.json','').replace('.geojson','').replace('.','_')
        
        # save results to variable named the same as the filename, excluding file extension
        exec(f"{var}=df")
